# -*- coding: utf-8 -*-
"""
Created on Mon Aug  5 14:39:02 2024

@author: Jonathan
"""
import sys

from qtpy import QtCore, QtWidgets

from .GUI_instances import clearAddDevice
from .icons import icons
from ..drivers import (list_drivers, load_driver_lib, get_connection_names,
                       get_driver_class, get_connection_class, get_class_args)
from ..config import get_all_devices_configs, save_config


class AddDeviceWindow(QtWidgets.QMainWindow):

    def __init__(self, parent: QtWidgets.QMainWindow = None):

        super().__init__(parent)
        self.mainGui = parent
        self.setWindowTitle('AUTOLAB - Add Device')
        self.setWindowIcon(icons['autolab'])

        self.statusBar = self.statusBar()

        self._prev_name = ''
        self._prev_conn = ''

        try:
            import pyvisa as visa
            self.rm = visa.ResourceManager()
        except:
            self.rm = None

        # Main layout creation
        layoutWindow = QtWidgets.QVBoxLayout()
        layoutWindow.setAlignment(QtCore.Qt.AlignTop)

        centralWidget = QtWidgets.QWidget()
        centralWidget.setLayout(layoutWindow)
        self.setCentralWidget(centralWidget)

        # Device nickname
        layoutDeviceNickname = QtWidgets.QHBoxLayout()
        layoutWindow.addLayout(layoutDeviceNickname)

        label = QtWidgets.QLabel('Device')

        self.deviceNickname = QtWidgets.QLineEdit()
        self.deviceNickname.setText('my_device')

        layoutDeviceNickname.addWidget(label)
        layoutDeviceNickname.addWidget(self.deviceNickname)

        # Driver name
        layoutDriverName = QtWidgets.QHBoxLayout()
        layoutWindow.addLayout(layoutDriverName)

        label = QtWidgets.QLabel('Driver')

        self.driversComboBox = QtWidgets.QComboBox()
        self.driversComboBox.addItems(list_drivers())
        self.driversComboBox.activated.connect(self.driverChanged)
        self.driversComboBox.setSizeAdjustPolicy(
            QtWidgets.QComboBox.AdjustToContents)

        layoutDriverName.addWidget(label)
        layoutDriverName.addStretch()
        layoutDriverName.addWidget(self.driversComboBox)

        # Driver connection
        layoutDriverConnection = QtWidgets.QHBoxLayout()
        layoutWindow.addLayout(layoutDriverConnection)

        label = QtWidgets.QLabel('Connection')

        self.connectionComboBox = QtWidgets.QComboBox()
        self.connectionComboBox.activated.connect(self.connectionChanged)
        self.connectionComboBox.setSizeAdjustPolicy(
            QtWidgets.QComboBox.AdjustToContents)

        layoutDriverConnection.addWidget(label)
        layoutDriverConnection.addStretch()
        layoutDriverConnection.addWidget(self.connectionComboBox)

        # Driver arguments
        self.layoutDriverArgs = QtWidgets.QHBoxLayout()
        layoutWindow.addLayout(self.layoutDriverArgs)

        self.layoutDriverOtherArgs = QtWidgets.QHBoxLayout()
        layoutWindow.addLayout(self.layoutDriverOtherArgs)

        # layout for optional args
        self.layoutOptionalArg = QtWidgets.QVBoxLayout()
        layoutWindow.addLayout(self.layoutOptionalArg)

        # Add argument
        layoutButtonArg = QtWidgets.QHBoxLayout()
        layoutWindow.addLayout(layoutButtonArg)

        addOptionalArg = QtWidgets.QPushButton('Add argument')
        addOptionalArg.setIcon(icons['add'])
        addOptionalArg.clicked.connect(lambda state: self.addOptionalArgClicked())

        layoutButtonArg.addWidget(addOptionalArg)
        layoutButtonArg.setAlignment(QtCore.Qt.AlignLeft)

        # Add device
        layoutButton = QtWidgets.QHBoxLayout()
        layoutWindow.addLayout(layoutButton)

        self.addButton = QtWidgets.QPushButton('Add Device')
        self.addButton.clicked.connect(self.addButtonClicked)

        layoutButton.addWidget(self.addButton)

        # update driver name combobox
        self.driverChanged()

        self.resize(self.minimumSizeHint())

    def addOptionalArgClicked(self, key: str = None, val: str = None):
        """ Add new layout for optional argument """
        layout = QtWidgets.QHBoxLayout()
        self.layoutOptionalArg.addLayout(layout)

        widget = QtWidgets.QLineEdit()
        widget.setText(key)
        layout.addWidget(widget)
        widget = QtWidgets.QLineEdit()
        widget.setText(val)
        layout.addWidget(widget)
        widget = QtWidgets.QPushButton()
        widget.setIcon(icons['remove'])
        widget.clicked.connect(lambda: self.removeOptionalArgClicked(layout))
        layout.addWidget(widget)

    def removeOptionalArgClicked(self, layout):
        """ Remove optional argument layout """
        for j in reversed(range(layout.count())):
            layout.itemAt(j).widget().setParent(None)
        layout.setParent(None)

    def addButtonClicked(self):
        """ Add the device to the config file """
        device_name = self.deviceNickname.text()
        driver_name = self.driversComboBox.currentText()
        conn = self.connectionComboBox.currentText()

        if device_name == '':
            self.setStatus('Need device name', 10000, False)
            return None

        device_dict = {}
        device_dict['driver'] = driver_name
        device_dict['connection'] = conn

        for layout in (self.layoutDriverArgs, self.layoutDriverOtherArgs):
            for i in range(0, (layout.count()//2)*2, 2):
                key = layout.itemAt(i).widget().text()
                val = layout.itemAt(i+1).widget().text()
                device_dict[key] = val

        for i in range(self.layoutOptionalArg.count()):
            layout = self.layoutOptionalArg.itemAt(i).layout()
            key = layout.itemAt(0).widget().text()
            val = layout.itemAt(1).widget().text()
            device_dict[key] = val

        # Update devices config
        device_config = get_all_devices_configs()
        new_device = {device_name: device_dict}
        device_config.update(new_device)
        save_config('devices_config', device_config)

        if hasattr(self.mainGui, 'initialize'): self.mainGui.initialize()

        self.close()

    def modify(self, nickname: str, conf: dict):
        """ Modify existing driver (not optimized) """

        self.setWindowTitle('AUTOLAB - Modify device')
        self.addButton.setText('Modify device')

        self.deviceNickname.setText(nickname)
        self.deviceNickname.setEnabled(False)
        driver_name = conf.pop('driver')
        conn = conf.pop('connection')
        index = self.driversComboBox.findText(driver_name)
        self.driversComboBox.setCurrentIndex(index)
        self.driverChanged()

        try:
            driver_lib = load_driver_lib(driver_name)
        except: pass
        else:
            list_conn = get_connection_names(driver_lib)
            if conn not in list_conn:
                if list_conn:
                    self.setStatus(f"Connection {conn} not found, switch to {list_conn[0]}", 10000, False)
                    conn = list_conn[0]
                else:
                    self.setStatus(f"No connections available for driver '{driver_name}'", 10000, False)
                    conn =  ''

        index = self.connectionComboBox.findText(conn)
        self.connectionComboBox.setCurrentIndex(index)
        if index != -1:
            self.connectionChanged()

        # Used to remove default value
        try:
            driver_lib = load_driver_lib(driver_name)
            driver_class = get_driver_class(driver_lib)
            assert hasattr(driver_class, 'slot_config')
        except:
            slot_config = '<MODULE_NAME>'
        else:
            slot_config = f'{driver_class.slot_config}'

        # Update args
        for layout in (self.layoutDriverArgs, self.layoutDriverOtherArgs):
            for i in range(0, (layout.count()//2)*2, 2):
                key = layout.itemAt(i).widget().text()
                if key in conf:
                    layout.itemAt(i+1).widget().setText(conf[key])
                    conf.pop(key)

        # Update optional args
        for i in reversed(range(self.layoutOptionalArg.count())):
            layout = self.layoutOptionalArg.itemAt(i).layout()
            key = layout.itemAt(0).widget().text()
            val_tmp = layout.itemAt(1).widget().text()
            # Remove default value
            if ((key == 'slot1' and val_tmp == slot_config)
                    or (key == 'slot1_name' and val_tmp == 'my_<MODULE_NAME>')):
                for j in reversed(range(layout.count())):
                    layout.itemAt(j).widget().setParent(None)
                layout.setParent(None)
            elif key in conf:
                layout.itemAt(1).widget().setText(conf[key])
                conf.pop(key)

        # Add remaining optional args from config
        for key, val in conf.items():
            self.addOptionalArgClicked(key, val)

    def driverChanged(self):
        """ Update driver information """
        driver_name = self.driversComboBox.currentText()

        if driver_name == self._prev_name: return None
        if driver_name == '':
            self.setStatus(f"Can't load driver associated with {self.deviceNickname.text()}", 10000, False)
            return None
        self._prev_name = driver_name

        try:
            driver_lib = load_driver_lib(driver_name)
        except Exception as e:
            # If error with driver remove all layouts
            self.setStatus(f"Can't load {driver_name}: {e}", 10000, False)

            self.connectionComboBox.clear()

            for layout in (self.layoutDriverArgs, self.layoutDriverOtherArgs):
                for i in reversed(range(layout.count())):
                    layout.itemAt(i).widget().setParent(None)

            for i in reversed(range(self.layoutOptionalArg.count())):
                layout = self.layoutOptionalArg.itemAt(i).layout()
                for j in reversed(range(layout.count())):
                    layout.itemAt(j).widget().setParent(None)
                layout.setParent(None)

            return None

        self.setStatus('')

        # Update available connections
        connections = get_connection_names(driver_lib)
        self.connectionComboBox.clear()
        self.connectionComboBox.addItems(connections)

        # update selected connection information
        self._prev_conn = ''
        self.connectionChanged()

        # reset layoutDriverOtherArgs
        for i in reversed(range(self.layoutDriverOtherArgs.count())):
            self.layoutDriverOtherArgs.itemAt(i).widget().setParent(None)

        # used to skip doublon key
        conn = self.connectionComboBox.currentText()
        try:
            driver_instance = get_connection_class(driver_lib, conn)
        except:
            connection_args = {}
        else:
            connection_args = get_class_args(driver_instance)

        # populate layoutDriverOtherArgs
        driver_class = get_driver_class(driver_lib)
        other_args = get_class_args(driver_class)
        for key, val in other_args.items():
            if key in connection_args: continue
            widget = QtWidgets.QLabel()
            widget.setText(key)
            self.layoutDriverOtherArgs.addWidget(widget)
            widget = QtWidgets.QLineEdit()
            widget.setText(str(val))
            self.layoutDriverOtherArgs.addWidget(widget)

        # reset layoutOptionalArg
        for i in reversed(range(self.layoutOptionalArg.count())):
            layout = self.layoutOptionalArg.itemAt(i).layout()
            for j in reversed(range(layout.count())):
                layout.itemAt(j).widget().setParent(None)
            layout.setParent(None)

        # populate layoutOptionalArg
        if hasattr(driver_class, 'slot_config'):
            self.addOptionalArgClicked('slot1', f'{driver_class.slot_config}')
            self.addOptionalArgClicked('slot1_name', 'my_<MODULE_NAME>')

    def connectionChanged(self):
        """ Update connection information """
        conn = self.connectionComboBox.currentText()

        if conn == self._prev_conn: return None
        self._prev_conn = conn

        driver_name = self.driversComboBox.currentText()
        try:
            driver_lib = load_driver_lib(driver_name)
        except:
            return None

        connection_args = get_class_args(
            get_connection_class(driver_lib, conn))

        # reset layoutDriverArgs
        for i in reversed(range(self.layoutDriverArgs.count())):
            self.layoutDriverArgs.itemAt(i).widget().setParent(None)

        conn_widget = None
        # populate layoutDriverArgs
        for key, val in connection_args.items():
            widget = QtWidgets.QLabel()
            widget.setText(key)
            self.layoutDriverArgs.addWidget(widget)

            widget = QtWidgets.QLineEdit()
            widget.setText(str(val))
            self.layoutDriverArgs.addWidget(widget)

            if key == 'address':
                conn_widget = widget

        if self.rm is not None and conn == 'VISA':
            widget = QtWidgets.QComboBox()
            widget.clear()
            conn_list = ('Available connections',) + tuple(self.rm.list_resources())
            widget.addItems(conn_list)
            if conn_widget is not None:
                widget.activated.connect(
                    lambda item, conn_widget=conn_widget: conn_widget.setText(
                        widget.currentText()) if widget.currentText(
                            ) != 'Available connections' else conn_widget.text())
            self.layoutDriverArgs.addWidget(widget)

    def setStatus(self, message: str, timeout: int = 0, stdout: bool = True):
        """ Modify the message displayed in the status bar and add error message to logger """
        self.statusBar.showMessage(message, timeout)
        if not stdout: print(message, file=sys.stderr)

    def closeEvent(self, event):
        """ Does some steps before the window is really killed """
        clearAddDevice()

        if not self.mainGui:
            QtWidgets.QApplication.quit()  # close the monitor app



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 20 22:08:29 2019

@author: qchat
"""
from typing import Union
import os
import sys
import queue

from qtpy import QtCore, QtWidgets, uic

from .data import DataManager
from .figure import FigureManager
from .monitor import MonitorManager
from ..icons import icons
from ..GUI_utilities import get_font_size, setLineEditBackground
from ..GUI_instances import clearMonitor
from ...paths import PATHS
from ...utilities import SUPPORTED_EXTENSION
from ...elements import Variable as Variable_og
from ...variables import Variable


class Monitor(QtWidgets.QMainWindow):

    def __init__(self,
                 variable: Union[Variable, Variable_og],
                 has_parent: bool = False, **kwargs):
        self.has_parent = has_parent  # Only for closeEvent
        self.variable = variable
        self._font_size = get_font_size()

        # Configuration of the window
        super().__init__()
        ui_path = os.path.join(os.path.dirname(__file__), 'interface.ui')
        uic.loadUi(ui_path, self)
        self.setWindowTitle(f"AUTOLAB - Monitor: {self.variable.address()}")
        self.setWindowIcon(icons['monitor'])
        # Queue
        self.queue = queue.Queue()
        self.timer = QtCore.QTimer(self)
        self.timer.setInterval(33)  # 30fps
        self.timer.timeout.connect(self.sync)

        # Window length
        self.windowLength_lineEdit.setText('10')
        self.windowLength_lineEdit.returnPressed.connect(self.windowLengthChanged)
        self.windowLength_lineEdit.textEdited.connect(lambda: setLineEditBackground(
            self.windowLength_lineEdit, 'edited', self._font_size))
        setLineEditBackground(
            self.windowLength_lineEdit, 'synced', self._font_size)

        self.xlabel = 'Time(s)'  # Is changed to x if ndarray or dataframe
        self.ylabel = f'{self.variable.address()}'  # OPTIMIZE: could depend on 1D or 2D

        if self.variable.unit is not None:
            self.ylabel += f'({self.variable.unit})'

        # Delay
        self.delay_lineEdit.setText('0.01')
        self.delay_lineEdit.returnPressed.connect(self.delayChanged)
        self.delay_lineEdit.textEdited.connect(lambda: setLineEditBackground(
            self.delay_lineEdit, 'edited', self._font_size))
        setLineEditBackground(self.delay_lineEdit, 'synced', self._font_size)

        self.pauseButton.clicked.connect(self.pauseButtonClicked)
        self.saveButton.clicked.connect(self.saveButtonClicked)
        self.clearButton.clicked.connect(self.clearButtonClicked)
        self.mean_checkBox.clicked.connect(self.mean_checkBoxClicked)
        self.min_checkBox.clicked.connect(self.min_checkBoxClicked)
        self.max_checkBox.clicked.connect(self.max_checkBoxClicked)

        # Managers
        self.dataManager = DataManager(self)
        self.figureManager = FigureManager(self)
        self.monitorManager = MonitorManager(self)

        # Start
        self.windowLengthChanged()
        self.delayChanged()
        self.monitorManager.start()
        self.timer.start()

        # Use to pause monitor on scan start
        self.pause_on_scan = False
        self.start_on_scan = False
        if self.has_parent:
            self.pause_on_scan_checkBox.clicked.connect(
                self.pause_on_scan_checkBoxClicked)
            self.start_on_scan_checkBox.clicked.connect(
                self.start_on_scan_checkBoxClicked)
        else:
            self.pause_on_scan_checkBox.hide()
            self.start_on_scan_checkBox.hide()

        for splitter in (self.splitter, ):
            for i in range(splitter.count()):
                handle = splitter.handle(i)
                handle.setStyleSheet("background-color: #DDDDDD;")
                handle.installEventFilter(self)

        # Handle kwargs at init
        try:
            mean = kwargs.pop('mean', False)
            if mean:
                self.mean_checkBox.setChecked(True)
                self.mean_checkBoxClicked()
            min = kwargs.pop('min', False)
            if min:
                self.min_checkBox.setChecked(True)
                self.min_checkBoxClicked()
            max = kwargs.pop('max', False)
            if max:
                self.max_checkBox.setChecked(True)
                self.max_checkBoxClicked()
            delay = kwargs.pop('delay', None)
            if delay is not None:
                self.delay_lineEdit.setText(f'{delay:g}')
                self.delayChanged()
            length = kwargs.pop('length', None)
            if length is not None:
                self.windowLength_lineEdit.setText(f'{length:g}')
                self.windowLengthChanged()
        except: pass
        if len(kwargs) != 0:
            self.setStatus(f"These kwargs argument doesn't exists: {list(kwargs)}", 10000, False)

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.Enter:
            obj.setStyleSheet("background-color: #AAAAAA;")  # Hover color
        elif event.type() == QtCore.QEvent.Leave:
            obj.setStyleSheet("background-color: #DDDDDD;")  # Normal color
        return super().eventFilter(obj, event)

    def sync(self):
        """ This function updates the data and then the figure.
        Function called by the time """
        # Empty the queue
        count = 0
        while not self.queue.empty():
            self.dataManager.addPoint(self.queue.get())
            count += 1

        # Upload the plot if new data available
        if count > 0:
            xlist, ylist = self.dataManager.getData()
            self.figureManager.update(xlist, ylist)

    def pauseButtonClicked(self):
        """ This function pause or resume the monitoring """
        if self.monitorManager.isPaused():
            self.timer.start()
            self.pauseButton.setText('Pause')
            self.monitorManager.resume()
        else:
            self.timer.stop()
            self.pauseButton.setText('Resume')
            self.monitorManager.pause()
            self.sync()

    def saveButtonClicked(self):
        """ This function is called when the SAVE button is pressed, and launch the procedure
        to save both the data and the figure """
        # Make sure the monitoring is paused
        if not self.monitorManager.isPaused():
            self.pauseButtonClicked()

        # Ask the filename of the output data
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, caption="Save data", directory=os.path.join(
                PATHS['last_folder'],
                f'{self.variable.address()}_monitor.txt'),
            filter=SUPPORTED_EXTENSION)

        path = os.path.dirname(filename)
        # Save the given path for future, the data and the figure if the path provided is valid
        if path != '':
            PATHS['last_folder'] = path
            self.setStatus('Saving data...', 5000)

            try:
                self.dataManager.save(filename)
                self.figureManager.save(filename)
                self.setStatus(f'Data successfully saved in {filename}.', 5000)
            except Exception as e:
                self.setStatus(f'Error while saving data: {e}', 10000, False)

    def clearButtonClicked(self):
        """ This function clear the displayed data """
        self.dataManager.clear()
        self.figureManager.clear()

    def mean_checkBoxClicked(self):
        """ This function toggle the mean plot """
        if not self.mean_checkBox.isChecked():
            self.figureManager.plot_mean.setData([], [])

        xlist, ylist = self.dataManager.getData()

        if len(xlist) > 0: self.figureManager.update(xlist,ylist)

    def min_checkBoxClicked(self):
        """ This function toggle the min plot """
        if not self.min_checkBox.isChecked():
            self.figureManager.plot_min.setData([], [])
            self.figureManager.ymin = None

        xlist, ylist = self.dataManager.getData()

        if len(xlist) > 0: self.figureManager.update(xlist, ylist)

    def max_checkBoxClicked(self):
        """ This function toggle the max plot """
        if not self.max_checkBox.isChecked():
            self.figureManager.plot_max.setData([], [])
            self.figureManager.ymax = None

        xlist, ylist = self.dataManager.getData()

        if len(xlist) > 0: self.figureManager.update(xlist, ylist)

    def pause_on_scan_checkBoxClicked(self):
        """ Change pause_on_scan variable """
        self.pause_on_scan = self.pause_on_scan_checkBox.isChecked()

    def start_on_scan_checkBoxClicked(self):
        """ Change start_on_scan variable """
        self.start_on_scan = self.start_on_scan_checkBox.isChecked()

    def closeEvent(self, event):
        """ This function does some steps before the window is really killed """
        self.monitorManager.close()
        self.timer.stop()
        self.figureManager.fig.deleteLater()  # maybe not useful for monitor but was source of crash in scanner if didn't close
        self.figureManager.figMap.deleteLater()

        clearMonitor(self.variable)

        if not self.has_parent:
            import pyqtgraph as pg
            try:
                # Prevent 'RuntimeError: wrapped C/C++ object of type ViewBox has been deleted' when reloading gui
                for view in pg.ViewBox.AllViews.copy().keys():
                    pg.ViewBox.forgetView(id(view), view)
                    # OPTIMIZE: forget only view used in monitor/gui
                pg.ViewBox.quit()
            except: pass

        for children in self.findChildren(QtWidgets.QWidget):
            children.deleteLater()

        super().closeEvent(event)

        if not self.has_parent:
            QtWidgets.QApplication.quit()  # close the app

    def windowLengthChanged(self):
        """ This function start the update of the window length in the data manager
        when a changed has been detected """
        # Send the new value
        try:
            value = float(self.windowLength_lineEdit.text())
            assert value > 0
            self.dataManager.setWindowLength(value)
        except: pass

        # Rewrite the GUI with the current value
        self.updateWindowLengthGui()

    def delayChanged(self):
        """ This function start the update of the delay in the thread manager
        when a changed has been detected """
        # Send the new value
        try:
            value = float(self.delay_lineEdit.text())
            assert value >= 0
            self.monitorManager.setDelay(value)
        except: pass

        # Rewrite the GUI with the current value
        self.updateDelayGui()

    def updateWindowLengthGui(self):
        """ This function ask the current value of the window length in the data
        manager, and then update the GUI """
        value = self.dataManager.getWindowLength()
        self.windowLength_lineEdit.setText(f'{value:g}')
        setLineEditBackground(self.windowLength_lineEdit, 'synced', self._font_size)

    def updateDelayGui(self):
        """ This function ask the current value of the delay in the data
        manager, and then update the GUI """
        value = self.monitorManager.getDelay()
        self.delay_lineEdit.setText(f'{value:g}')
        setLineEditBackground(self.delay_lineEdit, 'synced', self._font_size)

    def setStatus(self, message: str, timeout: int = 0, stdout: bool = True):
        """ Modify the message displayed in the status bar and add error message to logger """
        self.statusBar.showMessage(message, timeout)
        if not stdout: print(message, file=sys.stderr)



# -*- coding: utf-8 -*-
"""
Created on Thu Apr 11 20:13:46 2024

@author: jonathan
"""

import re
import os
import sys
from typing import Tuple, List, Union, Callable
from collections import defaultdict
import inspect

import numpy as np
import pandas as pd
from qtpy import QtWidgets, QtCore, QtGui
import pyqtgraph as pg

from .icons import icons
from ..paths import PATHS
from ..config import get_GUI_config
from ..devices import DEVICES, get_element_by_address
from ..variables import has_eval, EVAL, VARIABLES
from ..utilities import SUPPORTED_EXTENSION

# Fixes pyqtgraph/issues/3018 for pg<=0.13.7 (before pyqtgraph/pull/3070)
from pyqtgraph.graphicsItems.PlotDataItem import PlotDataItem

if hasattr(PlotDataItem, '_fourierTransform'):

    _fourierTransform_bugged = PlotDataItem._fourierTransform

    def _fourierTransform_fixed(self, x, y):
        if len(x) == 1: return np.array([0]), abs(y)
        return _fourierTransform_bugged(self, x, y)

    PlotDataItem._fourierTransform = _fourierTransform_fixed


ONCE = False


def get_font_size() -> int:
    GUI_config = get_GUI_config()

    return (int(float(GUI_config['font_size']))
                 if GUI_config['font_size'] != 'default'
                 else QtWidgets.QApplication.instance().font().pointSize())


def setLineEditBackground(obj, state: str, font_size: int = None):
    """ Sets background color of a QLineEdit widget based on its editing state """
    if state == 'synced': color='#D2FFD2' # vert
    if state == 'edited': color='#FFE5AE' # orange

    if font_size is None:
        obj.setStyleSheet(
            "QLineEdit:enabled {background-color: %s; color: #000000;}" % (
                color))
    else:
        obj.setStyleSheet(
            "QLineEdit:enabled {background-color: %s; font-size: %ipt; color: #000000;}" % (
                color, font_size))


CHECK_ONCE = True


def qt_object_exists(QtObject) -> bool:
    """ Return True if object exists (not deleted).
    Check if use pyqt5, pyqt6, pyside2 or pyside6 to use correct implementation
    """
    global CHECK_ONCE
    QT_API = os.environ.get("QT_API")

    if not CHECK_ONCE: return True
    try:
        if QT_API == "pyqt5":
            import sip
            return not sip.isdeleted(QtObject)
        if QT_API == "pyqt6":
            from PyQt6 import sip
            return not sip.isdeleted(QtObject)
        if QT_API == "pyside2":
            import shiboken2
            return shiboken2.isValid(QtObject)
        if QT_API =="pyside6":
            import shiboken6
            return shiboken6.isValid(QtObject)
        raise ModuleNotFoundError(f"QT_API '{QT_API}' unknown")
    except ModuleNotFoundError as e:
        print(f"Warning: {e}. Skip check if Qt Object not deleted.")
        CHECK_ONCE = False
        return True


class MyGraphicsLayoutWidget(pg.GraphicsLayoutWidget):
    # OPTIMIZE: could merge with myImageView to only have one class handling both lines and images

    def __init__(self):
        super().__init__()

        self.img_active = False

        # for plotting 1D
        ax = self.addPlot()
        self.ax = ax

        ax.setLabel("bottom", ' ', **{'color': pg.getConfigOption("foreground"),
                                      'font-size': '12pt'})
        ax.setLabel("left", ' ', **{'color': pg.getConfigOption("foreground"),
                                    'font-size': '12pt'})

        # Set your custom font for both axes
        my_font = QtGui.QFont('Arial', 12)
        my_font_tick = QtGui.QFont('Arial', 10)
        ax.getAxis("bottom").label.setFont(my_font)
        ax.getAxis("left").label.setFont(my_font)
        ax.getAxis("bottom").setTickFont(my_font_tick)
        ax.getAxis("left").setTickFont(my_font_tick)
        ax.showGrid(x=True, y=True)

        vb = ax.getViewBox()
        vb.enableAutoRange(enable=True)
        vb.setBorder(pg.mkPen(color=pg.getConfigOption("foreground")))

        ## Text label for the data coordinates of the mouse pointer
        dataLabel = pg.LabelItem(color=pg.getConfigOption("foreground"),
                                 parent=ax.getAxis('bottom'))
        dataLabel.anchor(itemPos=(1,1), parentPos=(1,1), offset=(0,0))

        def mouseMoved(point):
            """ This function marks the position of the cursor in data coordinates"""
            vb = ax.getViewBox()
            mousePoint = vb.mapSceneToView(point)
            l = f'x = {mousePoint.x():g},  y = {mousePoint.y():g}'
            dataLabel.setText(l)

        # data reader signal connection
        ax.scene().sigMouseMoved.connect(mouseMoved)

    def activate_img(self):
        """ Enable image feature """
        global ONCE
        # (py 3.6 -> pg 0.11.1, py 3.7 -> 0.12.4, py 3.8 -> 0.13.3, py 3.9 -> 0.13.7 (latest))
        # Disabled 2D plot if don't have pyqtgraph > 0.11
        pgv = pg.__version__.split('.')
        if int(pgv[0]) == 0 and int(pgv[1]) < 12:
            ONCE = True
            print("Can't use 2D plot for scan, need pyqtgraph >= 0.13.2", file=sys.stderr)
            # OPTIMIZE: could use ImageView instead?
            return None

        if not self.img_active:
            self.img_active = True

            # for plotting 2D
            img = pg.PColorMeshItem()
            self.ax.addItem(img)
            self.img = img

            # for plotting 2D colorbar
            if hasattr(self.ax, 'addColorBar') and hasattr(img, 'setLevels'):
                self.colorbar = self.ax.addColorBar(img, colorMap='viridis')  # pg 0.12.4
            else:
                if hasattr(pg, 'ColorBarItem'):
                    self.colorbar = pg.ColorBarItem(colorMap='viridis')  # pg 0.12.2
                else:
                    self.colorbar = pg.HistogramLUTItem()  # pg 0.11.0 (disabled)
                self.addItem(self.colorbar)

                if not ONCE:
                    ONCE = True
                    print('Skip colorbar update, need pyqtgraph >= 0.13.2', file=sys.stderr)

            self.colorbar.hide()

    def update_img(self, x, y, z):
        """ Update pcolormesh image """
        global ONCE
        z_no_nan = z[~np.isnan(z)]
        z[np.isnan(z)] = z_no_nan.min()-1e99  # OPTIMIZE: nan gives error, would prefer not to display empty values

        # Expand x and y arrays to define edges of the grid
        if len(x) == 1:
            x = np.append(x, x[-1]+1e-99)  # OPTIMIZE: first line too small and not visible if autoscale disabled, could use next x value instead but figure should not be aware of scan
        else:
            x = np.append(x, x[-1] + (x[-1] - x[-2]))

        if len(y) == 1:
            y = np.append(y, y[-1]+1e-99)
        else:
            y = np.append(y, y[-1] + (y[-1] - y[-2]))

        xv, yv = np.meshgrid(y, x)

        img = pg.PColorMeshItem()
        img.edgecolors = None
        img.setData(xv, yv, z.T)
        # OPTIMIZE: Changing log scale doesn't display correct axes
        if hasattr(img, 'setLevels'):  # pg 0.13.2 introduces setLevels in PColorMeshItem (py 3.8)
            self.colorbar.setImageItem(img)
        else:
            if not ONCE:
                ONCE = True
                print('Skip colorbar update, need pyqtgraph >= 0.13.2', file=sys.stderr)

        if isinstance(self.colorbar, pg.HistogramLUTItem):  # old
            self.colorbar.setLevels(z_no_nan.min(), z_no_nan.max())
        else:  # new
            self.colorbar.setLevels((z_no_nan.min(), z_no_nan.max()))

        # remove previous img and add new one (can't just refresh -> error if setData with nan and diff shape)
        self.ax.removeItem(self.img)
        self.img = img
        self.ax.addItem(self.img)

    def dragLeaveEvent(self, event):
        # Pyside6 triggers a leave event resuling in error:
        # QGraphicsView::dragLeaveEvent: drag leave received before drag enter
        pass
        # super().dragLeaveEvent(event)



def pyqtgraph_fig_ax() -> Tuple[MyGraphicsLayoutWidget, pg.PlotItem]:
    """ Return a formated fig and ax pyqtgraph for a basic plot """
    fig = MyGraphicsLayoutWidget()
    return fig, fig.ax


class myImageView(pg.ImageView):
    ''' Wrap of pg.ImageView with additionnal functionalities '''
    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        # update tick background on gradient change
        self.ui.histogram.gradient.sigGradientChanged.connect(self.update_ticks)

        self.figLineROI, self.axLineROI = pyqtgraph_fig_ax()
        self.figLineROI.hide()

        # update slice when change frame number in scanner
        self.timeLine.sigPositionChanged.connect(lambda: self.updateLineROI())

        # Create a checkbox for aspect ratio
        self.aspect_ratio_checkbox = QtWidgets.QCheckBox("Lock aspect ratio")
        self.aspect_ratio_checkbox.setChecked(True)
        self.aspect_ratio_checkbox.stateChanged.connect(self.toggle_aspect_ratio)

        # Store shapes in a list with visibility and remove information
        self.shapes = []
        self.shapes_plot = []

        # Create menu bar
        shape_toolbutton = QtWidgets.QToolButton(self)
        shape_toolbutton.setText(" Custom ROI ")
        shape_toolbutton.setPopupMode(QtWidgets.QToolButton.InstantPopup)

        menu_bar = QtWidgets.QMenu(self)
        menu_bar.aboutToShow.connect(self.update_shape_properties_menu)
        shape_toolbutton.setMenu(menu_bar)

        add_menu = menu_bar.addMenu("Add")

        add_line_action = QtWidgets.QAction("Line", add_menu)
        add_line_action.triggered.connect(self.add_line)
        add_menu.addAction(add_line_action)

        add_rectangle_action = QtWidgets.QAction("Rectangle", add_menu)
        add_rectangle_action.triggered.connect(self.add_rectangle)
        add_menu.addAction(add_rectangle_action)

        add_ellipse_action = QtWidgets.QAction("Ellipse", add_menu)
        add_ellipse_action.triggered.connect(self.add_ellipse)
        add_menu.addAction(add_ellipse_action)

        # Create a container for shape visibility toggles and delete buttons
        self.shape_properties_menu = menu_bar.addMenu("Properties")

        horizontalLayoutButton = QtWidgets.QHBoxLayout()
        horizontalLayoutButton.setSpacing(0)
        horizontalLayoutButton.setContentsMargins(0,0,0,0)
        horizontalLayoutButton.addWidget(self.aspect_ratio_checkbox)
        horizontalLayoutButton.addStretch()
        horizontalLayoutButton.addWidget(shape_toolbutton)

        widgetButton = QtWidgets.QWidget()
        widgetButton.setLayout(horizontalLayoutButton)

        verticalLayoutImageButton = QtWidgets.QVBoxLayout()
        verticalLayoutImageButton.setSpacing(0)
        verticalLayoutImageButton.setContentsMargins(0,0,0,0)
        verticalLayoutImageButton.addWidget(self)
        verticalLayoutImageButton.addWidget(widgetButton)

        widgetImageButton = QtWidgets.QWidget()
        widgetImageButton.setLayout(verticalLayoutImageButton)

        splitter = QtWidgets.QSplitter()
        splitter.setOrientation(QtCore.Qt.Vertical)
        splitter.addWidget(widgetImageButton)
        splitter.addWidget(self.figLineROI)
        splitter.setSizes([500,500])

        verticalLayoutMain = QtWidgets.QVBoxLayout()
        verticalLayoutMain.setSpacing(0)
        verticalLayoutMain.setContentsMargins(0,0,0,0)
        verticalLayoutMain.addWidget(splitter)

        centralWidget = QtWidgets.QWidget()
        centralWidget.setLayout(verticalLayoutMain)
        self.centralWidget = centralWidget

        for splitter in (splitter, ):
            for i in range(splitter.count()):
                handle = splitter.handle(i)
                handle.setStyleSheet("background-color: #DDDDDD;")
                handle.installEventFilter(self)

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.Enter:
            obj.setStyleSheet("background-color: #AAAAAA;")  # Hover color
        elif event.type() == QtCore.QEvent.Leave:
            obj.setStyleSheet("background-color: #DDDDDD;")  # Normal color
        return super().eventFilter(obj, event)

    def update_ticks(self):
        for tick in self.ui.histogram.gradient.ticks:
            tick.pen = pg.mkPen(pg.getConfigOption("foreground"))
            tick.currentPen = tick.pen
            tick.hoverPen = pg.mkPen(200, 120, 0)

    def toggle_aspect_ratio(self):
        self.getView().setAspectLocked(
            self.aspect_ratio_checkbox.isChecked())

    def show(self):
        self.centralWidget.show()

    def hide(self):
        self.centralWidget.hide()

    def roiChanged(self):
        pg.ImageView.roiChanged(self)
        for c in self.roiCurves:
            c.setPen(pg.getConfigOption("foreground"))

    def setImage(self, *args, **kwargs):
        pg.ImageView.setImage(self, *args, **kwargs)
        self.updateLineROI()

    def updateLineROI(self, shape: pg.ROI = None):
        shapes_to_plot = self.shapes_plot if shape is None else [shape]

        img = self.image if self.image.ndim == 2 else self.image[self.currentIndex]
        img = np.array([img])
        x,y = (0, 1) if self.imageItem.axisOrder == 'col-major' else (1, 0)

        for shape in shapes_to_plot:
            if shape not in self.shapes_plot:
                continue

            # OPTIMIZE: This behavior might not be ideal in all case, some user would prefer to have the average over the full range instead of the x line.
            if isinstance(shape, pg.LineSegmentROI):
                d2 = shape.getArrayRegion(img, self.imageItem, axes=(x+1, y+1))
                data = d2[0]
            elif isinstance(shape, pg.RectROI):
                d2 = shape.getArrayRegion(img, self.imageItem, axes=(y+1, x+1))
                data = np.mean(d2[0], axis=0)
            elif isinstance(shape, pg.EllipseROI):
                d2 = shape.getArrayRegion(img, self.imageItem, axes=(y+1, x+1))
                data = np.mean(d2[0], axis=0)
            else:
                d2 = shape.getArrayRegion(img, self.imageItem, axes=(x+1, y+1))
                data = d2[0]

            shape.plot.setData(data)

        self.figLineROI.setVisible(len(self.shapes_plot) != 0)

    def on_roi_clicked(self, roi: pg.ROI, ev: QtCore.QEvent):
        if ev.button() == QtCore.Qt.RightButton:
            roi.menu.actions_list['visible'].setChecked(roi.isVisible())
            global_pos = self.mapToGlobal(roi.getSceneHandlePositions()[0][1].toPoint())
            _ = roi.menu.exec_(global_pos)
            ev.accept()

    def toggle_all_plot(self, state: bool):
        for shape in self.shapes:
            if state:
                self.show_plot(shape)
            else:
                self.hide_plot(shape)

    def show_all_shapes(self):
        for shape in self.shapes:
            self.show_plot(shape)

    def hide_all_shapes(self):
        for shape in self.shapes:
            self.hide_plot(shape)

    def toggle_plot(self, shape: pg.ROI, state: bool):
        if state:
            self.show_plot(shape)
        else:
            self.hide_plot(shape)

    def show_plot(self, shape: pg.ROI):
        self.shapes_plot.append(shape)
        shape.plot.show()
        shape.menu.actions_list['plot'].setChecked(True)
        self.updateLineROI(shape)

    def hide_plot(self, shape: pg.ROI):
        self.shapes_plot.remove(shape)
        shape.plot.hide()
        shape.menu.actions_list['plot'].setChecked(False)
        self.figLineROI.setVisible(len(self.shapes_plot) != 0)

    def add_shape(self, shape: pg.ROI):

        if isinstance(shape, pg.LineSegmentROI):
            basename = 'Line'
        elif isinstance(shape, pg.RectROI):
            basename = 'Rectangle'
        elif isinstance(shape, pg.EllipseROI):
            basename = 'Ellipse'
        else:
            basename = 'Unknown'

        name = basename
        names = [shape.name for shape in self.shapes]
        compt = 0
        while True:
            if name in names:
                compt += 1
                name = basename + '_' + str(compt)
            else:
                break

        # Define shape menu
        shape_menu = QtWidgets.QMenu()
        shape_menu.setTitle(name)

        shape_actions = {}
        shape_menu.actions_list = shape_actions

        visibility_action = shape_menu.addAction("Show")
        visibility_action.setCheckable(True)
        visibility_action.setChecked(shape.isVisible())
        visibility_action.triggered.connect(shape.setVisible)
        shape_actions['visible'] = visibility_action

        plot_action = shape_menu.addAction("Plot")
        plot_action.setCheckable(True)
        plot_action.triggered.connect(lambda state: self.toggle_plot(shape, state))
        shape_actions['plot'] = plot_action

        delete_action = shape_menu.addAction("Remove")
        delete_action.triggered.connect(lambda: self.delete_shape(shape))
        shape_actions['delete'] = delete_action

        # Add attributes to ROI
        shape.name = name
        shape.menu = shape_menu
        shape.setAcceptedMouseButtons(QtCore.Qt.RightButton)
        shape.sigClicked.connect(self.on_roi_clicked)  # Connect sigClicked to the handler
        shape.sigRegionChanged.connect(lambda: self.updateLineROI(shape))
        shape.plot = self.axLineROI.plot([], [], pen=shape.pen)
        shape.plot.hide()

        self.addItem(shape)
        self.shapes.append(shape)
        self.shape_properties_menu.addMenu(shape.menu)

    def add_line(self):
        position_point = ((10, 10), (50, 10))  # Start and end position
        center_point = (0, 0)  # center point (x, y)

        line = pg.LineSegmentROI(position_point, center_point)
        line.setPen(pg.mkPen(QtGui.QColor(0, 255, 0), width=2))  # green

        self.add_shape(line)

    def add_rectangle(self):
        rect = pg.RectROI([10, 20], [30, 20])  # Position (x, y), size (width, height)
        rect.setPen(pg.mkPen(QtGui.QColor(255, 0, 0), width=2))  # red

        rect.addRotateHandle([0, 0], [0.5, 0.5])  # position at top-left corner, center at center of ROI

        self.add_shape(rect)

    def add_ellipse(self):
        ellipse = pg.EllipseROI([10, 50], [20, 20])  # Position (x, y), size (width, height)
        ellipse.setPen(pg.mkPen(QtGui.QColor(0, 0, 255), width=2))  # blue

        self.add_shape(ellipse)

    def delete_shape(self, shape: pg.ROI):
        """Delete a shape from the scene and update the shapes list and visibility menu."""
        self.shapes.remove(shape)
        self.getView().scene().removeItem(shape)
        if shape in self.shapes_plot:
            self.hide_plot(shape)
            self.axLineROI.removeItem(shape.plot)

    def delete_all_shapes(self):
        for shape in self.shapes.copy():
            self.delete_shape(shape)

    def toggle_all_visibility(self, checked: bool):
        for shape in self.shapes:
            shape.setVisible(checked)

    def update_shape_properties_menu(self):
        """Update the visibility menu to show only existing shapes."""
        self.shape_properties_menu.clear()  # Clear old entries

        all_menu = self.shape_properties_menu.addMenu('All')

        all_show = all_menu.addAction('Show')
        all_show.setCheckable(True)
        all_show.setChecked(all([shape.isVisible() for shape in self.shapes])
                                and len(self.shapes) != 0)
        all_show.triggered.connect(
            lambda: self.toggle_all_visibility(all_show.isChecked()))

        all_plot = all_menu.addAction('Plot')
        all_plot.setCheckable(True)
        all_plot.setChecked(len(self.shapes_plot) == len(self.shapes)
                                and len(self.shapes) != 0)
        all_plot.triggered.connect(
            lambda: self.toggle_all_plot(all_plot.isChecked()))

        all_delete = all_menu.addAction('Remove')
        all_delete.triggered.connect(self.delete_all_shapes)

        self.shape_properties_menu.addSeparator()

        for shape in self.shapes:
            self.shape_properties_menu.addMenu(shape.menu)

    def close(self):
        self.figLineROI.deleteLater()
        self.delete_all_shapes()
        super().close()


def pyqtgraph_image() -> Tuple[myImageView, QtWidgets.QWidget]:
    """ Return a formated ImageView and pyqtgraph widget for image plotting """
    imageView = myImageView()
    return imageView, imageView.centralWidget


class MyLineEdit(QtWidgets.QLineEdit):
    """https://stackoverflow.com/questions/28956693/pyqt5-qtextedit-auto-completion"""

    skip_has_eval = False
    use_variables = True
    use_devices = True
    use_np_pd = True
    completer: QtWidgets.QCompleter = None

    def __init__(self, *args):
        super().__init__(*args)

    def create_keywords(self) -> List[str]:
        """ Returns a list of all available keywords for completion """
        list_keywords = []

        if self.use_variables:
            list_keywords += list(VARIABLES)
            ## Could use this to see attributes of Variable like .raw .value
            # list_keywords += [f'{name}.{item}'
            #                     for name, var in VARIABLES.items()
            #                     for item in dir(var)
            #                     if not item.startswith('_') and not item.isupper()]
        if self.use_devices:
            list_keywords += [str(get_element_by_address(elements[0]).address())
                          for device in DEVICES.values()
                          if device.name not in list_keywords
                          for elements in device.get_structure()]

        if self.use_np_pd:
            if 'np' not in list_keywords:
                list_keywords += ['np']
                list_keywords += [f'np.{item}'
                                   for item in dir(np)
                                   if not item.startswith('_') and not item.isupper()]

            if 'pd' not in list_keywords:
                list_keywords += ['pd']
                list_keywords += [f'pd.{item}'
                                   for item in dir(pd)
                                   if not item.startswith('_') and not item.isupper()]
        return list_keywords

    def eventFilter(self, obj, event):
        """ Used when installEventFilter active """
        if (event.type() == QtCore.QEvent.KeyPress
                and event.key() == QtCore.Qt.Key_Tab):
            self.keyPressEvent(event)
            return True
        return super().eventFilter(obj, event)

    def setCompleter(self, completer: QtWidgets.QCompleter):
        """ Sets/removes completer """
        if self.completer:
            if self.completer.popup().isVisible(): return None
            self.completer.popup().close()
            try:
                self.completer.disconnect()  # PyQT
            except TypeError:
                self.completer.disconnect(self) # Pyside

        if not completer:
            self.removeEventFilter(self)
            self.completer = None
            return None

        self.installEventFilter(self)
        completer.setWidget(self)
        completer.setCompletionMode(QtWidgets.QCompleter.PopupCompletion)
        self.completer = completer
        self.completer.activated.connect(self.insertCompletion)

    def getCompletion(self) -> List[str]:
        model = self.completer.model()
        return model.stringList() if isinstance(
            model, QtCore.QStringListModel) else []

    def insertCompletion(self, completion: str, prefix: bool = True):
        cursor_pos = self.cursorPosition()
        text = self.text()

        prefix_length = (len(self.completer.completionPrefix())
                         if (prefix and self.completer) else 0)

        # Replace the current word with the completion
        new_text = (text[:cursor_pos - prefix_length]
                    + completion
                    + text[cursor_pos:])
        self.setText(new_text)
        self.setCursorPosition(cursor_pos - prefix_length + len(completion))

    def textUnderCursor(self) -> str:
        text = self.text()
        cursor_pos = self.cursorPosition()
        start = text.rfind(' ', 0, cursor_pos) + 1
        return text[start:cursor_pos]

    def focusInEvent(self, event):
        if self.completer:
            self.completer.setWidget(self)
        super().focusInEvent(event)

    def keyPressEvent(self, event):
        controlPressed = event.modifiers() == QtCore.Qt.ControlModifier
        tabPressed = event.key() == QtCore.Qt.Key_Tab
        specialTabPressed = tabPressed and controlPressed
        enterPressed = event.key() in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return)
        specialEnterPressed = enterPressed and controlPressed

        if specialTabPressed:
            self.insertCompletion('\t', prefix=False)
            return None

        if specialEnterPressed:
            self.insertCompletion('\n', prefix=False)
            return None

        # Fixe issue if press control after an eval (issue appears if do self.completer = None)
        if self.completer and controlPressed:
            super().keyPressEvent(event)
            return None

        if not self.completer or not tabPressed:
            if (self.completer and enterPressed and self.completer.popup().isVisible()):
                self.completer.activated.emit(
                    self.completer.popup().currentIndex().data())
            else:
                super().keyPressEvent(event)

        if self.skip_has_eval or has_eval(self.text()):
            if not self.completer:
                self.setCompleter(QtWidgets.QCompleter(self.create_keywords()))
        else:
            if self.completer:
                self.setCompleter(None)

        if self.completer and not tabPressed:
            self.completer.popup().close()

        if not self.completer or not tabPressed:
            return None

        completion_prefix = self.format_completion_prefix(self.textUnderCursor())

        new_keywords = self.create_new_keywords(
            self.create_keywords(), completion_prefix)
        keywords = self.getCompletion()

        if new_keywords != keywords:
            keywords = new_keywords
            self.setCompleter(QtWidgets.QCompleter(keywords))

        if completion_prefix != self.completer.completionPrefix():
            self.completer.setCompletionPrefix(completion_prefix)
            self.completer.popup().setCurrentIndex(
                self.completer.completionModel().index(0, 0))

        if self.completer.completionModel().rowCount() == 1:
            self.completer.setCompletionMode(
                QtWidgets.QCompleter.InlineCompletion)
            self.completer.complete()

            self.completer.activated.emit(self.completer.currentCompletion())
            self.completer.setCompletionMode(
                QtWidgets.QCompleter.PopupCompletion)
        else:
            cr = self.cursorRect()
            cr.setWidth(self.completer.popup().sizeHintForColumn(0)
                        + self.completer.popup().verticalScrollBar().sizeHint().width())
            self.completer.complete(cr)

    @staticmethod
    def format_completion_prefix(completion_prefix: str) -> str:
        """ Returns a simplified prefix for completion """
        if has_eval(completion_prefix):
            completion_prefix = completion_prefix[len(EVAL):]

        pattern = r'[a-zA-Z_][a-zA-Z0-9_]*(?:\.[a-zA-Z_][a-zA-Z0-9_]*)*'
        temp = [var for var in re.findall(pattern, completion_prefix)]

        if len(temp) > 0:
            position = completion_prefix.rfind(temp[-1])
            if position != -1:
                completion_prefix = completion_prefix[position:]

        special_char = ('(', '[', ',', ':', '-', '+', '^', '*', '/', '|')
        if completion_prefix.endswith(special_char):
            completion_prefix = ''

        return completion_prefix

    @staticmethod
    def create_new_keywords(list_keywords: List[str],
                            completion_prefix: str) -> List[str]:
        """ Returns a list with all available keywords and possible decomposition """
        # Create ordered list with all attributes and sub attributes
        master_list = []
        master_list.append(list_keywords)
        list_temp = list_keywords
        while len(list_temp) > 1:
            list_temp = list(set(
                [var[:-len(var.split('.')[-1])-1]
                 for var in list_temp if len(var.split('.')) != 0]))
            if '' in list_temp: list_temp.remove('')
            master_list.append(list_temp)

        # Filter attributes that contained completion_prefix and remove doublons
        flat_list = list(set(item
                             for sublist in master_list
                             for item in sublist
                             if item.startswith(completion_prefix)))

        # Group items by the number of dots
        dot_groups = defaultdict(list)
        for item in flat_list:
            dot_count = item.count('.')
            dot_groups[dot_count].append(item)

        # Sort the groups by the number of dots
        sorted_groups = sorted(
            dot_groups.items(), key=lambda x: x[0], reverse=True)

        # Extract items from each group and return as a list with sorted sublist
        sorted_list = [sorted(group) for _, group in sorted_groups]

        # Create list of all available keywords and possible decomposition
        new_keywords = []
        good = False
        for level in reversed(sorted_list):
            for item in level:
                if completion_prefix in item:
                    new_keywords.append(item)
                    good = True
            if good: break

        return new_keywords


class MyInputDialog(QtWidgets.QDialog):

    def __init__(self, parent: QtWidgets.QMainWindow, name: str):

        super().__init__(parent)
        self.setWindowTitle(name)

        lineEdit = MyLineEdit()
        lineEdit.setMaxLength(10000000)
        self.lineEdit = lineEdit

        # Add OK and Cancel buttons
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
            self)

        # Connect buttons
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(QtWidgets.QLabel(f"Set {name} value"))
        layout.addWidget(lineEdit)
        layout.addWidget(button_box)
        layout.addStretch()
        layout.setContentsMargins(10, 5, 10, 10)

        self.textValue = lineEdit.text
        self.setTextValue = lineEdit.setText
        self.resize(self.minimumSizeHint())

    def showEvent(self, event):
        """Focus and select the text in the lineEdit."""
        super().showEvent(event)
        self.lineEdit.setFocus()
        self.lineEdit.selectAll()

    def closeEvent(self, event):
        for children in self.findChildren(QtWidgets.QWidget):
            children.deleteLater()
        super().closeEvent(event)


class MyFileDialog(QtWidgets.QDialog):

    def __init__(self, parent: QtWidgets.QMainWindow, name: str,
                 mode: QtWidgets.QFileDialog):

        super().__init__(parent)
        if mode == QtWidgets.QFileDialog.AcceptOpen:
            self.setWindowTitle(f"Open file - {name}")
        elif mode == QtWidgets.QFileDialog.AcceptSave:
            self.setWindowTitle(f"Save file - {name}")

        file_dialog = QtWidgets.QFileDialog(self, QtCore.Qt.Widget)
        file_dialog.setAcceptMode(mode)
        file_dialog.setOption(QtWidgets.QFileDialog.DontUseNativeDialog)
        file_dialog.setWindowFlags(file_dialog.windowFlags() & ~QtCore.Qt.Dialog)
        file_dialog.setDirectory(PATHS['last_folder'])
        file_dialog.setNameFilters(SUPPORTED_EXTENSION.split(";;"))

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(file_dialog)
        layout.addStretch()
        layout.setSpacing(0)
        layout.setContentsMargins(0,0,0,0)

        self.exec_ = file_dialog.exec_
        self.selectedFiles = file_dialog.selectedFiles

    def closeEvent(self, event):
        for children in self.findChildren(QtWidgets.QWidget):
            children.deleteLater()

        super().closeEvent(event)


class MyQCheckBox(QtWidgets.QCheckBox):

    def __init__(self, parent):
        self.parent = parent
        super().__init__()

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        self.parent.valueEdited()
        try:
            inspect.signature(self.parent.write)
        except ValueError: pass  # For built-in method (occurs for boolean for action parameter)
        else:
            self.parent.write()


class MyQComboBox(QtWidgets.QComboBox):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.readonly = False
        self.wheel = True
        self.key = True

    def mousePressEvent(self, event):
        if not self.readonly:
            super().mousePressEvent(event)

    def keyPressEvent(self, event):
        if not self.readonly and self.key:
            super().keyPressEvent(event)

    def wheelEvent(self, event):
        if not self.readonly and self.wheel:
            super().wheelEvent(event)


class CustomMenu(QtWidgets.QMenu):
    """ Menu with action containing sub-menu for recipe and parameter selection """
    def __init__(self,
                 gui: QtWidgets.QMainWindow,
                 main_combobox: QtWidgets.QComboBox = None,
                 second_combobox: QtWidgets.QComboBox = None,
                 update_gui: Callable[[None], None] = None):

        super().__init__()
        self.gui = gui  #  gui is scanner
        self.main_combobox = main_combobox
        self.second_combobox = second_combobox
        self.update_gui = update_gui

        self.current_menu = 1
        self.selected_action = None

        self.recipe_names = [self.main_combobox.itemText(i)
                             for i in range(self.main_combobox.count())
                             ] if self.main_combobox else []

        self.HAS_RECIPE = len(self.recipe_names) > 1

        self.HAS_PARAM = (self.second_combobox.count() > 1
                          if self.second_combobox
                          else False)

        # To only show parameter if only one recipe
        if not self.HAS_RECIPE and self.HAS_PARAM:
            self.main_combobox = second_combobox
            self.second_combobox = None

            self.recipe_names = [self.main_combobox.itemText(i)
                                 for i in range(self.main_combobox.count())
                                 ] if self.main_combobox else []


    def addAnyAction(self, action_text='', icon_name='',
                     param_menu_active=False) -> Union[QtWidgets.QWidgetAction,
                                                       QtWidgets.QAction]:

        if self.HAS_RECIPE or (self.HAS_PARAM and param_menu_active):
            action = self.addCustomAction(action_text, icon_name,
                                          param_menu_active=param_menu_active)
        else:
            action = self.addAction(action_text)
            if icon_name != '':
                action.setIcon(icons[icon_name])

        return action

    def addCustomAction(self, action_text='', icon_name='',
                        param_menu_active=False) -> QtWidgets.QWidgetAction:
        """ Create an action with a sub menu for selecting a recipe and parameter """

        def close_menu():
            self.selected_action = action_widget
            self.close()

        def handle_hover():
            """ Fixe bad hover behavior and refresh radio_button """
            self.setActiveAction(action_widget)
            recipe_name = self.main_combobox.currentText()
            action = recipe_menu.actions()[self.recipe_names.index(recipe_name)]
            radio_button = action.defaultWidget()

            if not radio_button.isChecked():
                radio_button.setChecked(True)

        def handle_radio_click(name):
            """ Update parameters available, open parameter menu if available
            and close main menu to validate the action """
            if self.current_menu == 1:
                self.main_combobox.setCurrentIndex(self.recipe_names.index(name))
                if self.update_gui:
                    self.update_gui()
                recipe_menu.close()

                if param_menu_active:
                    param_items = [self.second_combobox.itemText(i)
                                   for i in range(self.second_combobox.count())
                                   ] if self.second_combobox else []

                    if len(param_items) > 1:
                        self.current_menu = 2
                        setup_menu_parameter(param_menu)
                        return None
            else:
                update_parameter(name)
                param_menu.close()
                self.current_menu = 1
                action_button.setMenu(recipe_menu)

            close_menu()

        def reset_menu(button: QtWidgets.QToolButton):
            QtWidgets.QApplication.sendEvent(
                button, QtCore.QEvent(QtCore.QEvent.Leave))
            self.current_menu = 1
            action_button.setMenu(recipe_menu)

        def setup_menu_parameter(param_menu: QtWidgets.QMenu):
            param_items = [self.second_combobox.itemText(i)
                           for i in range(self.second_combobox.count())]
            param_name = self.second_combobox.currentText()

            param_menu.clear()
            for param_name_i in param_items:
                add_radio_button_to_menu(param_name_i, param_name, param_menu)

            action_button.setMenu(param_menu)
            action_button.showMenu()

        def update_parameter(name: str):
            param_items = [self.second_combobox.itemText(i)
                           for i in range(self.second_combobox.count())]
            self.second_combobox.setCurrentIndex(param_items.index(name))
            if self.update_gui:
                self.update_gui()

        def add_radio_button_to_menu(item_name: str, current_name: str,
                                     target_menu: QtWidgets.QMenu):
            widget = QtWidgets.QFrame()
            radio_button = QtWidgets.QRadioButton(item_name, widget)
            action = QtWidgets.QWidgetAction(self.gui)
            action.setDefaultWidget(radio_button)
            target_menu.addAction(action)

            if item_name == current_name:
                radio_button.setChecked(True)

            radio_button.clicked.connect(
                lambda: handle_radio_click(item_name))

        # Add custom action
        action_button = QtWidgets.QToolButton()
        action_button.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)
        action_button.setText(f"   {action_text}")
        action_button.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        action_button.setAutoRaise(True)
        action_button.clicked.connect(close_menu)
        action_button.enterEvent = lambda event: handle_hover()
        if icon_name != '':
            action_button.setIcon(icons[icon_name])

        action_widget = QtWidgets.QWidgetAction(action_button)
        action_widget.setDefaultWidget(action_button)
        self.addAction(action_widget)

        recipe_menu = QtWidgets.QMenu()
        # recipe_menu.aboutToShow.connect(lambda: self.set_clickable(False))
        recipe_menu.aboutToHide.connect(lambda: reset_menu(action_button))

        if param_menu_active:
            param_menu = QtWidgets.QMenu()
            # param_menu.aboutToShow.connect(lambda: self.set_clickable(False))
            param_menu.aboutToHide.connect(lambda: reset_menu(action_button))

        recipe_name = self.main_combobox.currentText()

        for recipe_name_i in self.recipe_names:
            add_radio_button_to_menu(recipe_name_i, recipe_name, recipe_menu)

        action_button.setMenu(recipe_menu)

        return action_widget


class RecipeMenu(CustomMenu):

    def __init__(self, gui: QtWidgets.QMainWindow):  # gui is scanner

        main_combobox = gui.selectRecipe_comboBox if gui else None
        second_combobox = gui.selectParameter_comboBox if gui else None
        update_gui = gui._updateSelectParameter if gui else None

        super().__init__(gui, main_combobox, second_combobox, update_gui)



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:13:45 2019

@author: qchat
"""

import math as m

import numpy as np
from qtpy import QtCore, QtWidgets, QtGui

from .display import DisplayValues
from .customWidgets import ParameterQFrame
from ..GUI_utilities import (get_font_size, setLineEditBackground, MyLineEdit,
                             MyQComboBox, qt_object_exists)
from ..icons import icons
from ...utilities import clean_string, str_to_array, array_to_str, create_array
from ...variables import has_eval, has_variable, eval_safely


class ParameterManager:
    """ Manage the parameter from a recipe in a scan """

    def __init__(self, gui: QtWidgets.QMainWindow,
                 recipe_name: str, param_name: str):

        self.gui = gui
        self.recipe_name = recipe_name
        self.param_name = param_name

        self.point_or_step = "point"

        self._font_size = get_font_size()

        self.init_ui()

        # Do refresh at start
        self.refresh()

    def init_ui(self):
        # Parameter frame
        self.mainFrame = ParameterQFrame(
            self.gui, self.recipe_name, self.param_name)
        self.mainFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.mainFrame.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.mainFrame.customContextMenuRequested.connect(self.rightClick)
        self.mainFrame.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)

        # Parameter layout
        parameterLayout = QtWidgets.QVBoxLayout(self.mainFrame)
        parameterLayout.setContentsMargins(0,0,0,0)
        parameterLayout.setSpacing(0)

        frameParameter = QtWidgets.QFrame()
        frameParameter.setToolTip(
            "Drag and drop a variable or use the right click option of a " \
            "variable from the control panel to add a recipe to the tree: " \
            f"{self.recipe_name}")

        frameScanRange = QtWidgets.QFrame()

        parameterLayout.addWidget(frameParameter)
        parameterLayout.addWidget(frameScanRange)

        ## 1st row frame: Parameter
        ### Name
        self.parameterName_lineEdit = QtWidgets.QLineEdit('')
        self.parameterName_lineEdit.setSizePolicy(
            QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed))
        self.parameterName_lineEdit.setAlignment(QtCore.Qt.AlignCenter)
        self.parameterName_lineEdit.setToolTip(
            'Name of the parameter, as it will displayed in the data')
        self.parameterName_lineEdit.textEdited.connect(
            lambda: setLineEditBackground(
                self.parameterName_lineEdit, 'edited', self._font_size))
        self.parameterName_lineEdit.returnPressed.connect(self.nameChanged)
        self.parameterName_lineEdit.setEnabled(False)

        ### Address
        parameterAddressIndicator_label = QtWidgets.QLabel("Address:")
        parameterAddressIndicator_label.setAlignment(
            QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.parameterAddress_label = QtWidgets.QLabel("<variable>")
        self.parameterAddress_label.setAlignment(QtCore.Qt.AlignCenter)
        self.parameterAddress_label.setToolTip('Address of the parameter')

        ### Unit
        self.unit_label = QtWidgets.QLabel("uA")
        self.unit_label.setAlignment(
            QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)

        ### displayParameter button
        self.displayParameter = DisplayValues("Parameter", size=(250, 400))
        self.displayParameter.setWindowIcon(icons['ndarray'])
        self.displayParameter_pushButton = QtWidgets.QPushButton("Parameter")
        self.displayParameter_pushButton.setIcon(icons['ndarray'])
        self.displayParameter_pushButton.clicked.connect(
            self.displayParameterButtonClicked)

        ## 1sr row layout: Parameter
        layoutParameter = QtWidgets.QHBoxLayout(frameParameter)
        layoutParameter.addWidget(self.parameterName_lineEdit)
        layoutParameter.addWidget(self.unit_label)
        layoutParameter.addWidget(parameterAddressIndicator_label)
        layoutParameter.addWidget(self.parameterAddress_label)
        layoutParameter.addWidget(self.displayParameter_pushButton)

        ## 2nd row frame: Range
        self.frameScanRange_linLog = QtWidgets.QFrame()

        ### first grid widgets: start, stop
        labelStart = QtWidgets.QLabel("Start", self.frameScanRange_linLog)
        self.start_lineEdit = QtWidgets.QLineEdit('0', self.frameScanRange_linLog)
        self.start_lineEdit.setToolTip('Start value of the scan')
        self.start_lineEdit.setAlignment(QtCore.Qt.AlignCenter)

        labelEnd = QtWidgets.QLabel("End", self.frameScanRange_linLog)
        self.end_lineEdit = QtWidgets.QLineEdit('10', self.frameScanRange_linLog)
        self.end_lineEdit.setToolTip('End value of the scan')
        self.end_lineEdit.setAlignment(QtCore.Qt.AlignCenter)

        ### first grid layout: start, stop
        startEndGridLayout = QtWidgets.QGridLayout(self.frameScanRange_linLog)
        startEndGridLayout.addWidget(labelStart, 0, 0)
        startEndGridLayout.addWidget(self.start_lineEdit, 0, 1)
        startEndGridLayout.addWidget(labelEnd, 1, 0)
        startEndGridLayout.addWidget(self.end_lineEdit, 1, 1)

        startEndGridWidget = QtWidgets.QFrame(self.frameScanRange_linLog)
        startEndGridWidget.setLayout(startEndGridLayout)

        ### second grid widgets: mean, width
        labelMean = QtWidgets.QLabel("Mean", self.frameScanRange_linLog)
        self.mean_lineEdit = QtWidgets.QLineEdit('5', self.frameScanRange_linLog)
        self.mean_lineEdit.setAlignment(QtCore.Qt.AlignCenter)

        labelWidth = QtWidgets.QLabel("Width", self.frameScanRange_linLog)
        self.width_lineEdit = QtWidgets.QLineEdit('10', self.frameScanRange_linLog)
        self.width_lineEdit.setAlignment(QtCore.Qt.AlignCenter)

        ### second grid layout: mean, width
        meanWidthGridLayout = QtWidgets.QGridLayout(self.frameScanRange_linLog)
        meanWidthGridLayout.addWidget(labelMean, 0, 0)
        meanWidthGridLayout.addWidget(self.mean_lineEdit, 0, 1)
        meanWidthGridLayout.addWidget(labelWidth, 1, 0)
        meanWidthGridLayout.addWidget(self.width_lineEdit, 1, 1)

        meanWidthGridWidget = QtWidgets.QFrame(self.frameScanRange_linLog)
        meanWidthGridWidget.setLayout(meanWidthGridLayout)

        ### third grid widgets: npts, step, log
        labelNbpts = QtWidgets.QLabel("Nb points", self.frameScanRange_linLog)
        self.nbpts_lineEdit = QtWidgets.QLineEdit('11', self.frameScanRange_linLog)
        self.nbpts_lineEdit.setAlignment(QtCore.Qt.AlignCenter)

        self.labelStep = QtWidgets.QLabel("Step", self.frameScanRange_linLog)
        self.step_lineEdit = QtWidgets.QLineEdit('1', self.frameScanRange_linLog)
        self.step_lineEdit.setAlignment(QtCore.Qt.AlignCenter)

        ### third grid layout: npts, step, log
        nptsStepGridLayout = QtWidgets.QGridLayout(self.frameScanRange_linLog)
        nptsStepGridLayout.addWidget(labelNbpts, 0, 0)
        nptsStepGridLayout.addWidget(self.nbpts_lineEdit, 0, 1)
        nptsStepGridLayout.addWidget(self.labelStep, 1, 0)
        nptsStepGridLayout.addWidget(self.step_lineEdit, 1, 1)

        nptsStepGridWidget = QtWidgets.QFrame(self.frameScanRange_linLog)
        nptsStepGridWidget.setLayout(nptsStepGridLayout)

        ## 2nd row layout: Range
        layoutScanRange = QtWidgets.QHBoxLayout(self.frameScanRange_linLog)
        layoutScanRange.setContentsMargins(0,0,0,0)
        layoutScanRange.setSpacing(0)
        layoutScanRange.addWidget(startEndGridWidget)
        layoutScanRange.addStretch()
        layoutScanRange.addWidget(meanWidthGridWidget)
        layoutScanRange.addStretch()
        layoutScanRange.addWidget(nptsStepGridWidget)

        ## 2nd row bis frame: Values (hidden at start)
        self.frameScanRange_values = QtWidgets.QFrame()

        ### first grid widgets: values (hidden at start)
        labelValues = QtWidgets.QLabel("Values", self.frameScanRange_values)
        self.values_lineEdit = MyLineEdit('[0,1,2,3]', self.frameScanRange_values)
        self.values_lineEdit.setToolTip('Values of the scan')
        self.values_lineEdit.setAlignment(QtCore.Qt.AlignCenter)
        self.values_lineEdit.setMaxLength(10000000)
        # TODO: keep eval in values, show evaluated in evaluated
        self.labelEvaluatedValues = QtWidgets.QLabel(
            "Evaluated values", self.frameScanRange_values)
        self.evaluatedValues_lineEdit = QtWidgets.QLineEdit(
            '[0,1,2,3]', self.frameScanRange_values)
        self.evaluatedValues_lineEdit.setToolTip('Evaluated values of the scan')
        self.evaluatedValues_lineEdit.setAlignment(QtCore.Qt.AlignCenter)
        self.evaluatedValues_lineEdit.setMaxLength(10000000)
        self.evaluatedValues_lineEdit.setReadOnly(True)
        palette = self.evaluatedValues_lineEdit.palette()
        palette.setColor(
            QtGui.QPalette.Base, palette.color(QtGui.QPalette.Base).darker(107))
        self.evaluatedValues_lineEdit.setPalette(palette)

        ### first grid layout: values (hidden at start)
        valuesGridLayout = QtWidgets.QGridLayout(self.frameScanRange_values)
        valuesGridLayout.addWidget(labelValues, 0, 0)
        valuesGridLayout.addWidget(self.values_lineEdit, 0, 1)
        valuesGridLayout.addWidget(self.labelEvaluatedValues, 1, 0)
        valuesGridLayout.addWidget(self.evaluatedValues_lineEdit, 1, 1)

        valuesGridWidget = QtWidgets.QFrame(self.frameScanRange_values)
        valuesGridWidget.setLayout(valuesGridLayout)

        ## 2nd row bis layout: Values (hidden at start)
        layoutScanRange_values = QtWidgets.QHBoxLayout(self.frameScanRange_values)
        layoutScanRange_values.setContentsMargins(0,0,0,0)
        layoutScanRange_values.setSpacing(0)
        layoutScanRange_values.addWidget(valuesGridWidget)

        ## 3rd row frame: choice
        self.frameScanRange_choice = QtWidgets.QFrame()

        ### first grid widgets: choice
        self.comboBoxChoice = MyQComboBox(self.frameScanRange_choice)
        self.comboBoxChoice.wheel = False
        self.comboBoxChoice.key = False
        self.comboBoxChoice.addItems(['Linear', 'Log', 'Custom'])

        ### first grid layout: choice
        choiceGridLayout = QtWidgets.QGridLayout(self.frameScanRange_choice)
        choiceGridLayout.addWidget(self.comboBoxChoice, 0, 0)

        choiceGridWidget = QtWidgets.QFrame(self.frameScanRange_choice)
        choiceGridWidget.setLayout(choiceGridLayout)

        ## 3rd row layout: choice
        layoutScanRange_choice = QtWidgets.QHBoxLayout(self.frameScanRange_choice)
        layoutScanRange_choice.setContentsMargins(0,0,0,0)
        layoutScanRange_choice.setSpacing(0)
        layoutScanRange_choice.addWidget(choiceGridWidget)

        scanRangeLayout = QtWidgets.QHBoxLayout(frameScanRange)
        scanRangeLayout.setContentsMargins(0,0,0,0)
        scanRangeLayout.setSpacing(0)
        scanRangeLayout.addWidget(self.frameScanRange_linLog)
        scanRangeLayout.addWidget(self.frameScanRange_values)  # hidden at start
        scanRangeLayout.addWidget(self.frameScanRange_choice)

        # Widget 'return pressed' signal connections
        self.comboBoxChoice.activated.connect(self.scanRangeComboBoxChanged)
        self.nbpts_lineEdit.returnPressed.connect(self.nbptsChanged)
        self.step_lineEdit.returnPressed.connect(self.stepChanged)
        self.start_lineEdit.returnPressed.connect(self.startChanged)
        self.end_lineEdit.returnPressed.connect(self.endChanged)
        self.mean_lineEdit.returnPressed.connect(self.meanChanged)
        self.width_lineEdit.returnPressed.connect(self.widthChanged)
        self.values_lineEdit.returnPressed.connect(self.valuesChanged)

        # Widget 'text edited' signal connections
        self.nbpts_lineEdit.textEdited.connect(lambda: setLineEditBackground(
            self.nbpts_lineEdit,'edited', self._font_size))
        self.step_lineEdit.textEdited.connect(lambda: setLineEditBackground(
            self.step_lineEdit,'edited', self._font_size))
        self.start_lineEdit.textEdited.connect(lambda: setLineEditBackground(
            self.start_lineEdit,'edited', self._font_size))
        self.end_lineEdit.textEdited.connect(lambda: setLineEditBackground(
            self.end_lineEdit,'edited', self._font_size))
        self.mean_lineEdit.textEdited.connect(lambda: setLineEditBackground(
            self.mean_lineEdit,'edited', self._font_size))
        self.width_lineEdit.textEdited.connect(lambda: setLineEditBackground(
            self.width_lineEdit,'edited', self._font_size))
        self.values_lineEdit.textEdited.connect(lambda: setLineEditBackground(
            self.values_lineEdit,'edited', self._font_size))

    def _removeWidget(self):
        if hasattr(self, 'mainFrame'):
            try:
                self.mainFrame.hide()
                self.mainFrame.deleteLater()
                del self.mainFrame
            except: pass

    def changeName(self, newName):
        self.param_name = newName
        self.mainFrame.param_name = newName

    def refresh(self):
        """ Refreshes all the values of the name and address of the scan parameter,
        from the configuration center """
        # Parameter
        element = self.gui.configManager.getParameterElement(self.recipe_name, self.param_name)

        if element is not None:
            address = element.address()
            unit = element.unit
        else:
            address = self.param_name
            unit = ''

        self.parameterName_lineEdit.setEnabled(True)
        self.parameterName_lineEdit.setText(self.param_name)
        setLineEditBackground(self.parameterName_lineEdit, 'synced', self._font_size)
        self.parameterAddress_label.setText(address)
        if unit in ('', None):
            self.unit_label.setText('')
        else:
            self.unit_label.setText(f'({unit})')

        # Values if defined
        if self.gui.configManager.hasCustomValues(self.recipe_name, self.param_name):
            raw_values = self.gui.configManager.getValues(self.recipe_name, self.param_name)

            str_raw_values = raw_values if has_eval(
                raw_values) else array_to_str(raw_values)

            if not has_eval(raw_values):
                str_values = str_raw_values
                self.evaluatedValues_lineEdit.hide()
                self.labelEvaluatedValues.hide()
            else:
                values = eval_safely(raw_values)
                try: values = create_array(values)
                except: str_values = values
                else: str_values = array_to_str(values)
                self.evaluatedValues_lineEdit.show()
                self.labelEvaluatedValues.show()

            self.frameScanRange_linLog.hide()
            self.frameScanRange_values.show()
            self.values_lineEdit.setText(f'{str_raw_values}')
            self.evaluatedValues_lineEdit.setText(f'{str_values}')
            setLineEditBackground(self.values_lineEdit, 'synced', self._font_size)
            self.comboBoxChoice.setCurrentIndex(2)  # Custom
        else:
            self.frameScanRange_linLog.show()
            self.frameScanRange_values.hide()

            # Range
            xrange = self.gui.configManager.getRange(self.recipe_name, self.param_name)

            # Start
            start = xrange[0]
            self.start_lineEdit.setText(f'{start:g}')
            setLineEditBackground(self.start_lineEdit, 'synced', self._font_size)

            # End
            end = xrange[1]
            self.end_lineEdit.setText(f'{end:g}')
            setLineEditBackground(self.end_lineEdit, 'synced', self._font_size)

            # Mean
            mean = (start + end) / 2
            self.mean_lineEdit.setText(f'{mean:g}')
            setLineEditBackground(self.mean_lineEdit, 'synced', self._font_size)

            # Width
            width = abs(end - start)
            self.width_lineEdit.setText(f'{width:g}')
            setLineEditBackground(self.width_lineEdit, 'synced', self._font_size)

            # Nbpts
            nbpts = self.gui.configManager.getNbPts(self.recipe_name, self.param_name)
            step = self.gui.configManager.getStep(self.recipe_name, self.param_name)

            self.nbpts_lineEdit.setText(f'{nbpts:g}')
            setLineEditBackground(self.nbpts_lineEdit, 'synced', self._font_size)

            # Log
            log: bool = self.gui.configManager.getLog(self.recipe_name, self.param_name)

            # Step
            if log:
                self.comboBoxChoice.setCurrentIndex(1)  # Log
                self.labelStep.hide()
                self.step_lineEdit.hide()
                self.step_lineEdit.setText('')
            else:
                self.comboBoxChoice.setCurrentIndex(0)  # Linear
                self.step_lineEdit.setText(f'{step:g}')
                self.labelStep.show()
                self.step_lineEdit.show()

            setLineEditBackground(self.step_lineEdit, 'synced', self._font_size)

        if self.displayParameter.active:
            try: paramValues = self.gui.configManager.getParamDataFrame(
                self.recipe_name, self.param_name)
            except:
                self.gui.setStatus(
                    f"Wrong format for parameter '{self.param_name}'", 10000, False)
                return None

            str_values = array_to_str(paramValues[self.param_name].values)
            self.evaluatedValues_lineEdit.setText(f'{str_values}')
            self.displayParameter.refresh(paramValues)

    def displayParameterButtonClicked(self):
        """ Opens a table widget with the array of the corresponding parameter """
        if not self.displayParameter.active:
            try: paramValues = self.gui.configManager.getParamDataFrame(
                self.recipe_name, self.param_name)
            except Exception as e:
                self.gui.setStatus(f"Wrong format for parameter '{self.param_name}': {e}", 10000)
                return None
            str_values = array_to_str(paramValues[self.param_name].values)
            self.evaluatedValues_lineEdit.setText(f'{str_values}')
            self.displayParameter.refresh(paramValues)

        self.displayParameter.show()

    def nameChanged(self):
        """ Changes the name of the scan parameter """
        newName = self.parameterName_lineEdit.text()
        newName = clean_string(newName)

        if newName != '':
            self.gui.configManager.renameParameter(self.recipe_name, self.param_name, newName)

    def nbptsChanged(self):
        """ Changes the number of points of the parameter """
        value = self.nbpts_lineEdit.text()

        try:
            value = int(float(value))
            assert value > 0

            self.gui.configManager.setNbPts(self.recipe_name, self.param_name, value)
            self.point_or_step = "point"
        except:
            self.refresh()

    def stepChanged(self):
        """ Changes the step size of the parameter """
        value = self.step_lineEdit.text()

        try:
            if value == "inf" or float(value) == 0:
                self.nbpts_lineEdit.setText('1')
                self.nbptsChanged()
                return

            value = float(value)
            assert value > 0

            self.gui.configManager.setStep(self.recipe_name, self.param_name, value)

            self.point_or_step = "step"
        except:
            self.refresh()

    def startChanged(self):
        """ Changes the start value of the parameter """
        value = self.start_lineEdit.text()

        try:
            value = float(value)
            log: bool = self.gui.configManager.getLog(self.recipe_name, self.param_name)
            if log: assert value > 0

            xrange = list(self.gui.configManager.getRange(self.recipe_name, self.param_name))
            xrange[0] = value

            self.gui.configManager.setRange(self.recipe_name, self.param_name, xrange)
        except:
            self.refresh()

    def endChanged(self):
        """ Changes the end value of the parameter """
        value = self.end_lineEdit.text()

        try:
            value = float(value)
            log:bool = self.gui.configManager.getLog(self.recipe_name, self.param_name)
            if log: assert value > 0
            xrange = list(self.gui.configManager.getRange(self.recipe_name, self.param_name))
            xrange[1] = value

            self.gui.configManager.setRange(self.recipe_name, self.param_name, xrange)
        except:
            self.refresh()

    def meanChanged(self):
        """ Changes the mean value of the parameter """
        value = self.mean_lineEdit.text()

        try:
            value = float(value)
            log: bool = self.gui.configManager.getLog(self.recipe_name, self.param_name)
            if log: assert value > 0
            xrange = list(self.gui.configManager.getRange(self.recipe_name, self.param_name))
            xrange_new = xrange.copy()
            xrange_new[0] = value - (xrange[1] - xrange[0])/2
            xrange_new[1] = value + (xrange[1] - xrange[0])/2
            if log:
                assert xrange_new[0] > 0
                assert xrange_new[1] > 0

            self.gui.configManager.setRange(self.recipe_name, self.param_name, xrange_new)
        except:
            self.refresh()

    def widthChanged(self):
        """ Changes the width of the parameter """
        value = self.width_lineEdit.text()

        try:
            value = float(value)
            log: bool = self.gui.configManager.getLog(self.recipe_name, self.param_name)
            if log: assert value > 0
            xrange = list(self.gui.configManager.getRange(self.recipe_name, self.param_name))
            xrange_new = xrange.copy()
            xrange_new[0] = (xrange[1]+xrange[0])/2 - value/2
            xrange_new[1] = (xrange[1]+xrange[0])/2 + value/2
            if log:
                assert xrange_new[0] > 0
                assert xrange_new[1] > 0

            self.gui.configManager.setRange(self.recipe_name, self.param_name, xrange_new)
        except:
            self.refresh()

    def scanRangeComboBoxChanged(self):
        choice = self.comboBoxChoice.currentText()

        raw_values = self.gui.configManager.getValues(
            self.recipe_name, self.param_name)

        if choice == "Custom":
            self.gui.configManager.setValues(
                self.recipe_name, self.param_name, raw_values)
        else:
            CUSTOM_VALUES = self.gui.configManager.hasCustomValues(
                self.recipe_name, self.param_name)
            if CUSTOM_VALUES:
                param = self.gui.configManager.getParameter(
                    self.recipe_name, self.param_name)
                # OPTIMIZE: would be cleaner in configManager but easier that way
                param.pop('values')

                if isinstance(raw_values, np.ndarray): values = raw_values
                else:
                    try: values = str_to_array(self.evaluatedValues_lineEdit.text())
                    except: values = np.linspace(0, 10, 11)  # OPTIMIZE: should know last values when start scan (need to update values)

                param['range'] = (values[0], values[-1])

                if 'log' not in param: param['log'] = False

                self.gui.configManager.setNbPts(
                    self.recipe_name, self.param_name, len(values))

            if choice == "Linear": self.setLog(False)
            elif choice == "Log": self.setLog(True)

    def setLog(self, log: bool):
        """ Changes the log state of the parameter """
        if log:
            self.point_or_step = "point"
            xrange = list(self.gui.configManager.getRange(self.recipe_name, self.param_name))
            change = False

            if xrange[1] <= 0:
                xrange[1] = 1
                change = True

            if xrange[0] <= 0:
                xrange[0] = 10**(m.log10(xrange[1]) - 1)
                change = True

            if change: self.gui.configManager.setRange(self.recipe_name, self.param_name, xrange)

        self.gui.configManager.setLog(self.recipe_name, self.param_name, log)

    def valuesChanged(self):
        """ Changes values of the parameter """
        raw_values = self.values_lineEdit.text()

        try:
            if not has_eval(raw_values):
                raw_values = str_to_array(raw_values)
                assert len(raw_values) != 0, "Cannot have empty array"
                values = raw_values
            elif not has_variable(raw_values):
                values = eval_safely(raw_values)
                if not isinstance(values, str):
                    values = create_array(values)
                    assert len(values) != 0, "Cannot have empty array"

            self.gui.configManager.setValues(self.recipe_name, self.param_name, raw_values)
        except Exception as e:
            self.gui.setStatus(f"Wrong format for parameter '{self.param_name}': {e}", 10000)

    def rightClick(self, position: QtCore.QPoint):
        """ Provides the menu when the user right click on a parameter """
        menu = QtWidgets.QMenu()


        addAction = menu.addAction("Add parameter")
        addAction.setIcon(icons['add'])

        removeAction = menu.addAction(f"Remove {self.param_name}")
        removeAction.setIcon(icons['remove'])

        choice = menu.exec_(self.mainFrame.mapToGlobal(position))

        if choice == addAction:
            self.gui.configManager.addParameter(self.recipe_name)
        if choice == removeAction:
            self.gui.configManager.removeParameter(self.recipe_name, self.param_name)

    # PROCESSING STATE BACKGROUND
    ###########################################################################

    def setProcessingState(self, state: str):
        """ Sets the background color of the parameter address during the scan """
        if not qt_object_exists(self.parameterAddress_label):
            return None
        if state == 'idle':
            self.parameterAddress_label.setStyleSheet(
                f"font-size: {self._font_size}pt;")
        else:
            if state == 'started': color = '#ff8c1a'
            if state == 'finished': color = '#70db70'
            self.parameterAddress_label.setStyleSheet(
                f"background-color: {color}; font-size: {self._font_size}pt;")

    def close(self):
        """ Called by scanner on closing """
        self.displayParameter.close()



# -*- coding: utf-8 -*-
"""
Created on Fri Oct 18 23:09:51 2019

@author: qchat
"""
from typing import Any, List, Tuple
import re
import ast
from io import StringIO
import platform
import os

import numpy as np
import pandas as pd


SUPPORTED_EXTENSION = "Text Files (*.txt);; Supported text Files (*.txt;*.csv;*.dat);; Any Files (*)"


def emphasize(txt: str, sign: str = '-') -> str:
    ''' Returns:    ---
                    txt
                    ---
    '''
    return sign*len(txt) + '\n' + txt + '\n' + sign*len(txt)


def underline(txt: str, sign: str = '-') -> str:
    ''' Returns:
                    txt
                    ---
    '''
    return txt + '\n' + sign*len(txt)


def clean_string(txt: str) -> str:
    """ Returns txt without special characters """
    for character in r'*."/\[]:;|, ':
        txt = txt.replace(character, '')

    return txt


def two_columns(txt_list: List[str]) -> str:
    ''' Returns a string of the form:
        txt[0]                         txt[1]
        with a minimal spacing between the first character of txt1 and txt2 '''
    spacing = max(len(txt[0]) for txt in txt_list) + 5

    return '\n'.join([txt[0] + ' '*(spacing-len(txt[0])) + txt[1]
                          for txt in txt_list])


def boolean(value: Any) -> bool:
    """ Convert value from "True" or "False" or float, int, bool to bool """
    if value == "True": value = True
    elif value == "False": value = False
    else: value = bool(int(float(value)))

    return value


def str_to_value(s: str) -> Any:
    ''' Tries to convert string to int, float, bool or None in this order '''
    try:
        int_val = int(s)
        if str(int_val) == s: return int_val
    except ValueError: pass

    try:
        float_val = float(s)
        return float_val
    except ValueError: pass

    if s.lower() in ('true', 'false'):
        return s.lower() == 'true'

    if s == 'None': s = None
    # If none of the above works, return the string itself
    return s


def str_to_tuple(s: str) -> Tuple[List[str], int]:
    ''' Convert string to Tuple[List[str], int] '''
    e = "Input string does not match the required format Tuple[List[str], int]"
    try:
        result = ast.literal_eval(s)
        e = f"{result} does not match the required format Tuple[List[str], int]"
        assert (isinstance(result, (tuple, list))
                and len(result) == 2
                and isinstance(result[0], (list, tuple))
                and isinstance(result[1], int)), e
        result = ([str(res) for res in result[0]], result[1])
        return result
    except Exception:
        raise Exception(e)

def create_array(value: Any) -> np.ndarray:
    ''' Format an int, float, list or numpy array to a numpy array with at least
    one dimension '''
    # check validity of array, raise error if dtype not int or float
    np.array(value, ndmin=1, dtype=float)
    # Convert to ndarray and keep original dtype
    value = np.asarray(value)

    # want ndim >= 1 to avoid having float if 0D
    while value.ndim < 1:
        value = np.expand_dims(value, axis=0)
    return value


def str_to_array(s: str) -> np.ndarray:
    ''' Convert string to a numpy array '''
    if "," in s: ls = re.sub(r'\s,+', ',', s)
    else: ls = re.sub(r'\s+', ',', s)
    test = ast.literal_eval(ls)

    return create_array(test)


def array_to_str(value: Any, threshold: int = None, max_line_width: int = None) -> str:
    ''' Convert a numpy array to a string '''
    return np.array2string(np.array(value), separator=',', suppress_small=True,
                           threshold=threshold, max_line_width=max_line_width)


def str_to_dataframe(s: str) -> pd.DataFrame:
    ''' Convert a string to a pandas DataFrame '''
    if s == '\r\n':  # empty
        df = pd.DataFrame()
    else:
        value_io = StringIO(s)
        df = pd.read_csv(value_io, sep="\t")
    return df


def dataframe_to_str(value: pd.DataFrame, threshold=1000) -> str:
    ''' Convert a pandas DataFrame to a string '''
    if isinstance(value, str) and value == '': value = None
    return pd.DataFrame(value).head(threshold).to_csv(index=False, sep="\t")  # can't display full data to QLineEdit, need to truncate (numpy does the same)


def str_to_data(s: str) -> Any:
    """ Convert str to data with special format for ndarray and dataframe """
    if '\t' in s and '\n' in s:
        try: s = str_to_dataframe(s)
        except: pass
    elif '[' in s:
        try: s = str_to_array(s)
        except: pass
    else:
        try: s = str_to_value(s)
        except: pass
    return s


def data_to_str(value: Any) -> str:
    """ Convert data to str with special format for ndarray and dataframe """
    if isinstance(value, np.ndarray):
        raw_value_str = array_to_str(value, threshold=1000000, max_line_width=9000000)
    elif isinstance(value, pd.DataFrame):
        raw_value_str = dataframe_to_str(value, threshold=1000000)
    else:
        raw_value_str = str(value)
    return raw_value_str


def open_file(filename: str):
    ''' Opens a file using the platform specific command '''
    system = platform.system()
    if system == 'Windows': os.startfile(filename)
    elif system == 'Linux': os.system(f'gedit "{filename}"')
    elif system == 'Darwin': os.system(f'open "{filename}"')


def data_to_dataframe(data: Any) -> pd.DataFrame:
    """ Format data to DataFrame """
    try: data = pd.DataFrame(data)
    except ValueError: data = pd.DataFrame([data])

    data.columns = data.columns.astype(str)
    data_type = data.values.dtype

    try:
        data[data.columns] = data[data.columns].apply(pd.to_numeric, errors="coerce")
    except ValueError:
        pass  # OPTIMIZE: This happens when there is identical column name

    if len(data) != 0:
        assert not data.isnull().values.all(), f"Datatype '{data_type}' is not supported"
        if data.iloc[-1].isnull().values.all():  # if last line is full of nan, remove it
            data = data[:-1]

    if data.shape[1] == 1:
        data.rename(columns = {'0': '1'}, inplace=True)
        data.insert(0, "0", range(data.shape[0]))

    return data


def input_wrap(*args):
    """ Wrap input function to avoid crash with Spyder using Qtconsole=5.3 """
    input_allowed = True
    try:
        import spyder_kernels
        import qtconsole
    except ModuleNotFoundError:
        pass
    else:
        if hasattr(spyder_kernels, "console") and hasattr(qtconsole, "__version__"):
            if qtconsole.__version__.startswith("5.3"):
                print("Warning: Spyder crashes with input() if Qtconsole=5.3, skip user input.")
                input_allowed = False
    if input_allowed:
        ans = input(*args)
    else:
        ans = "yes"

    return ans



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:22:43 2019

@author: qchat
"""
from typing import Any, Tuple, List

import pandas as pd
import numpy as np
from qtpy import QtWidgets


class DataManager:

    def __init__(self, gui: QtWidgets.QMainWindow):

        self.gui = gui
        self.windowLength = 10
        self.xlist = []
        self.ylist = []

    def setWindowLength(self, value: float):
        """ This function set the value of the window length """
        self.windowLength = value

    def getWindowLength(self) -> float:
        """ This function returns the value of the window legnth """
        return self.windowLength

    def getData(self) -> Tuple[List, List]:
        """ This function update the data of the provided plot object """
        return self.xlist, self.ylist

    def save(self, filename: str):
        """ This function save the data in a file with the provided filename"""
        if self.xlist is not None:
            df = pd.DataFrame({self.gui.xlabel: self.xlist,
                               self.gui.ylabel: self.ylist})
            df.to_csv(filename, index=False)
        else: # Image
            df = pd.DataFrame(self.ylist)
            df.to_csv(filename, index=False, header=None)  # faster and handle better different dtype than np.savetxt

    def addPoint(self, point: Tuple[Any, Any]):
        """ This function either replace list by array or add point to list depending on datapoint type """
        y = point[1]

        if isinstance(y, (np.ndarray, pd.DataFrame)):
            if self.gui.windowLength_lineEdit.isVisible():
                self.gui.xlabel = 'x'
                self.gui.figureManager.setLabel('x', self.gui.xlabel)
                self.gui.windowLength_lineEdit.hide()
                self.gui.windowLength_label.hide()
                self.gui.dataDisplay.hide()

            if isinstance(y, np.ndarray):
                if len(y.T.shape) in (0, 1) or y.T.shape[0] == 2:
                    self._addArray(y.T)
                else:
                    self._addImage(y)  # Defined which axe is major using pg.setConfigOption('imageAxisOrder', 'row-major') in gui start-up so no need to .T data
            elif isinstance(y, pd.DataFrame):
                self._addArray(y.values.T)
        else:
            if not self.gui.windowLength_lineEdit.isVisible():
                self.gui.xlabel = 'Time(s)'
                self.gui.figureManager.setLabel('x', self.gui.xlabel)
                self.gui.windowLength_lineEdit.show()
                self.gui.windowLength_label.show()
                self.gui.dataDisplay.show()

            self._addPoint(point)

    def _addImage(self, image: np.ndarray):
        """ Add image to ylist data as np.ndarray """
        self.xlist = None
        self.ylist = image

    def _addArray(self, array: np.ndarray):
        """ This function replace an dataset [x,y] x is time y is array """
        if len(array.shape) == 0:
            y_array = array
            self.xlist = np.array([0])
            self.ylist = np.array([y_array])
        elif len(array.shape) == 1:
            y_array = array
            # Replace data
            self.xlist = np.arange(len(y_array))
            self.ylist = np.array(y_array)
        elif array.shape[0] >= 2:
            x_array = array[0]
            y_array = array[1] # OPTIMIZE: add button in gui to choose x and y like in scan

            # Replace data
            self.xlist = np.array(x_array)
            self.ylist = np.array(y_array)

    def _addPoint(self, point: Tuple[float, float]):
        """ This function append a datapoint [x,y] in the lists of data """
        if not hasattr(self.xlist, 'append'): self.clear()  # avoid error when switching from array to point

        x, y = point

        # Append data
        self.xlist.append(x)
        self.ylist.append(y)

        # Remove too old data (regarding the window length)
        while (max(self.xlist) - min(self.xlist)) > self.windowLength:
            self.xlist.pop(0)
            self.ylist.pop(0)

    def clear(self):
        try:
            self.xlist.clear()
            self.ylist.clear()
        except AttributeError:  # numpy array has no clear method
            self.xlist = []
            self.ylist = []



# -*- coding: utf-8 -*-
"""
Created on Fri Oct 18 22:20:14 2019

@author: qchat
"""

import os
import sys
import inspect
from typing import Type, Tuple, List, Any

import numpy as np
import pandas as pd

from .paths import PATHS
from .utilities import emphasize, clean_string, SUPPORTED_EXTENSION


class Element():

    def __init__(self, parent: Type, element_type: str, name: str):

        self.name = name
        self._element_type = element_type
        self._parent = parent
        self._help = None

    def address(self) -> str:
        """ Returns the address of the given element.
        <module.submodule.variable> """
        if self._parent is not None:
            return self._parent.address() + '.' + self.name
        return self.name


class Variable(Element):

    def __init__(self, parent: Type, config: dict):

        super().__init__(parent, 'variable', config['name'])

        # Type
        assert 'type' in config, f"Variable {self.address()}: Missing variable type"
        assert config['type'] in [int, float, bool, str, bytes, tuple, np.ndarray, pd.DataFrame], f"Variable {self.address()} configuration: Variable type not supported in autolab"
        self.type = config['type']
        if self.type in [tuple]:
            self.value = ([], -1)

        # Read and write function
        assert 'read' in config or 'write' in config, f"Variable {self.address()} configuration: no 'read' nor 'write' functions provided"

        # Read function
        self.read_function = None
        self.read_init = False
        # if config['type'] in [tuple]: assert 'read' in config, f"Variable {self.address()} configuration: Must provide a read function"
        if 'read' in config:
            assert inspect.ismethod(config['read']), f"Variable {self.address()} configuration: Read parameter must be a function"
            self.read_function = config['read']
            if 'read_init' in config:
                assert isinstance(config['read_init'], bool), f"Variable {self.address()} configuration: read_init parameter must be a boolean"
                self.read_init = bool(config['read_init'])

        # Write function
        self.write_function = None
        if 'write' in config:
            assert inspect.ismethod(config['write']), f"Variable {self.address()} configuration: Write parameter must be a function"
            self.write_function = config['write']

        # Unit
        self.unit = None
        if 'unit' in config:
            assert isinstance(config['unit'], str), f"Variable {self.address()} configuration: Unit parameter must be a string"
            self.unit = config['unit']

        # Help
        if 'help' in config:
            assert isinstance(config['help'], str), f"Variable {self.address()} configuration: Info parameter must be a string"
            self._help = config['help']

        # Properties
        self.writable = self.write_function is not None
        self.readable = self.read_function is not None
        self.numerical = self.type in [int, float]
        self.parameter_allowed = self.writable and self.numerical

        # Signals for GUI
        self._read_signal = None
        self._write_signal = None

    def save(self, path: str, value: Any = None):
        """ This function measure the variable and saves its value in the provided path """

        assert self.readable, f"The variable {self.address()} is not configured to be measurable"

        if os.path.isdir(path):
            path = os.path.join(path, self.address()+'.txt')

        if value is None: value = self() # New measure if value not provided

        if self.type in [int, float, bool, str, tuple]:
            with open(path, 'w') as f: f.write(str(value))
        elif self.type == bytes:
            with open(path, 'wb') as f: f.write(value)
        elif self.type == np.ndarray:
            try:
                value = pd.DataFrame(value)  # faster and handle better different dtype than np.savetxt
                value.to_csv(path, index=False, header=None)
            except:
                # Avoid error if strange ndim, 0 or (1,2,3) ... was occuring in GUI scan when doing $eval:[1] instead of $eval:np.array([1]). Now GUI forces array to ndim=1
                print(f"Warning, can't save {value}")
        elif self.type == pd.DataFrame:
            value.to_csv(path, index=False)
        else:
            raise ValueError("The variable {self.address()} of type {self.type} cannot be saved.")

    def help(self):
        """ This function prints informations for the user about the current variable """
        print(self)

    def __str__(self) -> str:
        """ This function returns informations for the user about the current variable """
        display = '\n' + emphasize(f'Variable {self.address()}') + '\n'
        if self._help is not None: display += f'Help: {self._help}\n'
        display += '\n'

        display += 'Readable: '
        if self.readable: display += f"YES (driver function '{self.read_function.__name__}')\n"
        else: display += 'NO\n'

        display += 'Writable: '
        if self.writable: display += f"YES (driver function '{self.write_function.__name__}')\n"
        else: display += 'NO\n'

        display += f'Type: {self.type.__name__}\n'

        display += 'Unit: '
        if self.unit is not None: display += f'{self.unit}\n'
        else: display += 'None\n'

        return display

    def __call__(self, value: Any = None) -> Any:
        """ Measure or set the value of the variable """
        # GET FUNCTION
        if value is None:
            assert self.readable, f"The variable {self.address()} is not readable"
            answer = self.read_function()
            if self._read_signal is not None: self._read_signal.emit_read(answer)
            if self.type in [tuple]:  # OPTIMIZE: could be generalized to any variable but fear could lead to memory issue
                self.value = answer
            return answer

        # SET FUNCTION
        assert self.writable, f"The variable {self.address()} is not writable"

        if isinstance(value, np.ndarray) or self.type in [np.ndarray]:
            value = np.array(value, ndmin=1)  # ndim=1 to avoid having float if 0D
        else:
            value = self.type(value)
        if self.type in [tuple]:  # OPTIMIZE: could be generalized to any variable but fear could lead to memory issue
            self.value = value
        self.write_function(value)
        if self._write_signal is not None: self._write_signal.emit_write(value)
        return None


class Action(Element):

    def __init__(self, parent: Type, config: dict):

        super().__init__(parent, 'action', config['name'])

        # Do function
        assert 'do' in config, f"Action {self.address()}: Missing 'do' function"
        assert inspect.ismethod(config['do']), f"Action {self.address()} configuration: Do parameter must be a function"
        self.function = config['do']

        # Argument
        self.type = None
        self.unit = None
        if 'param_type' in config:
            assert config['param_type'] in [int, float, bool, str, bytes, tuple, np.ndarray, pd.DataFrame], f"Action {self.address()} configuration: Argument type not supported in autolab"
            self.type = config['param_type']
            if 'param_unit' in config:
                assert isinstance(config['param_unit'], str), f"Action {self.address()} configuration: Argument unit parameter must be a string"
                self.unit = config['param_unit']

        # Help
        if 'help' in config:
            assert isinstance(config['help'], str), f"Action {self.address()} configuration: Info parameter must be a string"
            self._help = config['help']

        self.has_parameter = self.type is not None

        if self.type in [tuple]:
            self.value = ([], -1)

        # Signals for GUI
        self._write_signal = None

    def help(self):
        """ This function prints informations for the user about the current variable """
        print(self)

    def __str__(self) -> str:
        """ This function returns informations for the user about the current variable """
        display = '\n' + emphasize(f'Action {self.address()}') + '\n'
        if self._help is not None: display+=f'Help: {self._help}\n'
        display += '\n'

        display += f"Driver function: '{self.function.__name__}'\n"

        if self.has_parameter:
            display += f'Parameter: YES (type: {self.type.__name__})'
            if self.unit is not None: display += f'(unit: {self.unit})'
            display += '\n'
        else:
            display += 'Parameter: NO\n'

        return display

    def __call__(self, value: Any = None) -> Any:
        """ Executes the action """
        # DO FUNCTION
        assert self.function is not None, f"The action {self.address()} is not configured to be actionable"
        if self.has_parameter:
            if value is not None:
                if isinstance(value, np.ndarray):
                    value = np.array(value, ndmin=1)  # ndim=1 to avoid having float if 0D
                else:
                    value = self.type(value)
                self.function(value)
            elif self.unit in ('open-file', 'save-file', 'filename'):
                if self.unit == 'filename':  # LEGACY (may be removed later)
                    print(f"Using 'filename' as unit is depreciated in favor of 'open-file' and 'save-file'" \
                          f"\nUpdate driver '{self.address().split('.')[0]}' to remove this warning",
                          file=sys.stderr)
                    self.unit = 'open-file'

                from qtpy import QtWidgets
                _ = QtWidgets.QApplication(sys.argv)  # Needed if started outside of GUI

                if self.unit == 'open-file':
                    filename, _ = QtWidgets.QFileDialog.getOpenFileName(
                        caption=f"Open file - {self.address()}",
                        directory=PATHS['last_folder'],
                        filter=SUPPORTED_EXTENSION)
                elif self.unit == 'save-file':
                    filename, _ = QtWidgets.QFileDialog.getSaveFileName(
                        caption=f"Save file - {self.address()}",
                        directory=PATHS['last_folder'],
                        filter=SUPPORTED_EXTENSION)

                if filename != '':
                    path = os.path.dirname(filename)
                    PATHS['last_folder'] = path
                    value = filename
                    self.function(value)
                else:
                    print(f"Action '{self.address()}' cancel filename selection")

            elif self.unit == "user-input":

                from qtpy import QtWidgets
                _ = QtWidgets.QApplication(sys.argv)  # Needed if started outside of GUI
                # OPTIMIZE: dialog closes on instantiation inside Spyder
                response, _ = QtWidgets.QInputDialog.getText(
                    None, self.address(), f"Set {self.address()} value",
                    QtWidgets.QLineEdit.Normal)

                if response != '':
                    value = response
                    self.function(value)
            else:
                assert value is not None, f"The action {self.address()} requires an argument"
        else:
            assert value is None, f"The action {self.address()} doesn't require an argument"
            self.function()

        if self.type in [tuple]:  # OPTIMIZE: could be generalized to any variable but fear could lead to memory issue
            self.value = value
        if self._write_signal is not None: self._write_signal.emit_write(value)


class Module(Element):

    def __init__(self, parent: Type, config: dict):

        super().__init__(parent, 'module', config['name'])

        self._mod = {}
        self._var = {}
        self._act = {}
        self._read_init_list = []

        # Object - instance
        assert 'object' in config, f"Module {self.address()}: missing module object"
        self.instance = config['object']

        # Help
        if 'help' in config:
            assert isinstance(config['help'], str), f"Module {self.address()} configuration: Help parameter must be a string"
            self._help = config['help']

        # Loading instance
        assert hasattr(self.instance, 'get_driver_model'), "There is no function 'get_driver_model' in the driver class"
        driver_config = self.instance.get_driver_model()
        assert isinstance(driver_config, list), f"Module {self.address()} configuration: 'get_driver_model' output must be a list of dictionnaries"

        for config_line in driver_config:
            # General check
            assert isinstance(config_line, dict), f"Module {self.address()} configuration: 'get_driver_model' output must be a list of dictionnaries"

            # Name check
            assert 'name' in config_line, f"Module {self.address()} configuration: missing 'name' key in one dictionnary"
            assert isinstance(config_line['name'], str), f"Module {self.address()} configuration: elements names must be a string"
            name = clean_string(config_line['name'])
            assert name != '', f"Module {self.address()}: elements names cannot be empty"

            # Element type check
            assert 'element' in config_line, f"Module {self.address()}, Element {name} configuration: missing 'element' key in the dictionnary"
            assert isinstance(config_line['element'], str), f"Module {self.address()}, Element {name} configuration: element type must be a string"
            element_type = config_line['element']
            assert element_type in ['module', 'variable', 'action'], f"Module {self.address()}, Element {name} configuration: Element type has to be either 'module','variable' or 'action'"

            if element_type == 'module':
                # Check name uniqueness
                assert name not in self.get_names(), f"Module {self.address()}, Submodule {name} configuration: '{name}' already exists"
                self._mod[name] = Module(self, config_line)

            elif element_type == 'variable':
                # Check name uniqueness
                assert name not in self.get_names(), f"Module {self.address()}, Variable {name} configuration: '{name}' already exists"
                self._var[name] = Variable(self, config_line)
                if self._var[name].read_init:
                    self._read_init_list.append(self._var[name])

            elif element_type == 'action':
                # Check name uniqueness
                assert name not in self.get_names(), f"Module {self.address()}, Action {name} configuration: '{name}' already exists"
                self._act[name] = Action(self, config_line)

    def get_module(self, name: str) -> Type:  # -> Module
        """ Returns the submodule of the given name """
        assert name in self.list_modules(), f"The submodule '{name}' does not exist in module {self.address()}"
        return self._mod[name]

    def list_modules(self) -> List[str]:
        """ Returns a list with the names of all existing submodules """
        return list(self._mod)

    def get_variable(self, name: str) -> Variable:
        """ Returns the variable with the given name """
        assert name in self.list_variables(), f"The variable '{name}' does not exist in module {self.address()}"
        return self._var[name]

    def list_variables(self) -> List[str]:
        """ Returns a list with the names of all existing variables attached to this module """
        return list(self._var)

    def get_action(self, name) -> Action:
        """ Returns the action with the given name """
        assert name in self.list_actions(), f"The action '{name}' does not exist in device {self.address()}"
        return self._act[name]

    def list_actions(self) -> List[str]:
        """ Returns a list with the names of all existing actions attached to this module """
        return list(self._act)

    def get_names(self) -> List[str]:
        """ Returns the list of the names of all the elements of this module """
        return self.list_modules() + self.list_variables() + self.list_actions()

    def __getattr__(self, attr: str) -> Element:
        if attr in self.list_variables(): return self.get_variable(attr)
        if attr in self.list_actions(): return self.get_action(attr)
        if attr in self.list_modules(): return self.get_module(attr)
        raise AttributeError(f"'{attr}' not found in module '{self.address()}'")

    def get_structure(self) -> List[Tuple[str, str]]:
        """ Returns the structure of the module as a list containing each element address associated with its type as
        [['address1', 'variable'], ['address2', 'action'],...] """
        structure = []

        for mod in self.list_modules():
            structure += self.get_module(mod).get_structure()
        for var in self.list_variables():
            structure.append((self.get_variable(var).address(), 'variable'))
        for act in self.list_actions():
            structure.append((self.get_action(act).address(), 'action'))

        return structure

    def sub_hierarchy(self, level: int = 0) -> List[Tuple[str, str, int]]:
        ''' Returns a list of the sub hierarchy of this module
        [[name: str, type: str, level: int]]
        '''
        h = []

        from .devices import Device  # import here to avoid ImportError circular import
        if isinstance(self, Device): h.append((self.name, 'Device/Module', level))
        else: h.append((self.name, 'Module', level))

        for mod in self.list_modules():
            h += self.get_module(mod).sub_hierarchy(level+1)
        for var in self.list_variables():
            h.append((var, 'Variable', level+1))
        for act in self.list_actions():
            h.append((act, 'Action', level+1))

        return h

    def help(self):
        """ This function prints informations for the user about the availables
        submodules, variables and action attached to the current module """
        print(self)

    def __str__(self) -> str:
        """ This function returns informations for the user about the availables
        submodules, variables and action attached to the current module """
        display ='\n' + emphasize(f'Module {self.address()}') + '\n'

        if self._help is not None:
            display += f'Help: {self._help}\n'

        display += '\nElement hierarchy:'
        hierarchy = self.sub_hierarchy()

        for i, h_step in enumerate(hierarchy):
            if i == 0: prefix = ''
            else: prefix = '|- '
            txt = ' '*3*h_step[2] + f'{prefix}{h_step[0]}'
            display += '\n' + txt + ' '*(30 - len(txt)) + f'({h_step[1]})'

        return display

    def __dir__(self):
        """ For auto-completion """
        return (self.list_modules() + self.list_variables()
                + self.list_actions() + ['help', 'instance'])



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 25 10:46:14 2024

@author: jonathan
"""
import sys

from qtpy import QtWidgets, QtGui

from .icons import icons
from .GUI_instances import clearDriverInstaller
from ..paths import DRIVER_SOURCES, DRIVER_REPOSITORY
from ..repository import install_drivers, _download_driver, _get_drivers_list_from_github
from ..drivers import update_drivers_paths


class DriverInstaller(QtWidgets.QMainWindow):

    def __init__(self, parent=None):
        """ GUI to select which driver to install from the official github repo """

        super().__init__()

        self.setWindowTitle("AUTOLAB - Driver Installer")
        self.setWindowIcon(icons['autolab'])
        self.setFocus()
        self.activateWindow()

        self.statusBar = self.statusBar()

        official_folder = DRIVER_SOURCES['official']
        official_url = DRIVER_REPOSITORY[official_folder]

        list_driver = []
        try:
            list_driver = _get_drivers_list_from_github(official_url)
        except:
            self.setStatus(f'Warning: Cannot access {official_url}', 10000, False)

        self.mainGui = parent
        self.url = official_url  # TODO: use dict DRIVER_REPOSITORY to have all urls
        self.list_driver = list_driver
        self.OUTPUT_DIR = official_folder

        self.init_ui()

        self.adjustSize()

    def init_ui(self):
        centralWidget = QtWidgets.QWidget()
        self.setCentralWidget(centralWidget)

        # OFFICIAL DRIVERS
        formLayout = QtWidgets.QFormLayout()
        centralWidget.setLayout(formLayout)

        self.masterCheckBox = QtWidgets.QCheckBox(f"From {DRIVER_REPOSITORY[DRIVER_SOURCES['official']]}:")
        self.masterCheckBox.setChecked(False)
        self.masterCheckBox.stateChanged.connect(self.masterCheckBoxChanged)
        formLayout.addRow(self.masterCheckBox)

        # Init table size
        sti = QtGui.QStandardItemModel()
        for i in range(len(self.list_driver)):
            sti.appendRow([QtGui.QStandardItem(str())])

        # Create table
        tab = QtWidgets.QTableView()
        tab.setModel(sti)
        tab.verticalHeader().setVisible(False)
        tab.horizontalHeader().setVisible(False)
        tab.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        tab.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        tab.setAlternatingRowColors(True)
        tab.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                          QtWidgets.QSizePolicy.Expanding)
        tab.setSizeAdjustPolicy(
            QtWidgets.QAbstractScrollArea.AdjustToContents)

        if self.list_driver:  # OPTIMIZE: c++ crash if no driver in list!
            tab.horizontalHeader().setSectionResizeMode(
                0, QtWidgets.QHeaderView.ResizeToContents)

        # Init checkBox
        self.list_checkBox = []
        for i, driver_name in enumerate(self.list_driver):
            checkBox = QtWidgets.QCheckBox(f"{driver_name}")
            checkBox.setChecked(False)
            self.list_checkBox.append(checkBox)
            tab.setIndexWidget(sti.index(i, 0), checkBox)

        formLayout.addRow(QtWidgets.QLabel(""), tab)

        download_pushButton = QtWidgets.QPushButton()
        download_pushButton.clicked.connect(self.installListDriver)
        download_pushButton.setText("Download")
        formLayout.addRow(download_pushButton)

    def masterCheckBoxChanged(self):
        """ Checked all the checkBox related to the official github repo """
        state = self.masterCheckBox.isChecked()
        for checkBox in self.list_checkBox:
            checkBox.setChecked(state)

    def installListDriver(self):
        """ Install all the drivers for which the corresponding checkBox has been checked """
        list_bool = [
            checkBox.isChecked() for checkBox in self.list_checkBox]
        list_driver_to_download = [
            driver_name for (driver_name, driver_bool) in zip(
                self.list_driver, list_bool) if driver_bool]

        try:
            # Better for all drivers
            if all(list_bool):
                install_drivers(skip_input=True)
            # Better for several drivers
            elif any(list_bool):
                for driver_name in list_driver_to_download:
                    # self.setStatus(f"Downloading {driver_name}", 5000)  # OPTIMIZE: currently thread blocked by installer so don't show anything until the end
                    e = _download_driver(
                        self.url, driver_name, self.OUTPUT_DIR, _print=False)
                    if e is not None:
                        print(e, file=sys.stderr)
                        # self.setStatus(e, 10000, False)
        except Exception as e:
            self.setStatus(f'Error: {e}', 10000, False)
        else:
            self.setStatus('Finished!', 5000)

        # Update available drivers
        update_drivers_paths()

    def closeEvent(self, event):
        """ This function does some steps before the window is really killed """
        clearDriverInstaller()

        super().closeEvent(event)

        if not self.mainGui:
            QtWidgets.QApplication.quit()  # close the app

    def setStatus(self, message: str, timeout: int = 0, stdout: bool = True):
        """ Modify the message displayed in the status bar and add error message to logger """
        self.statusBar.showMessage(message, timeout)
        if not stdout: print(message, file=sys.stderr)



# -*- coding: utf-8 -*-
"""
Created on Tue Sep 17 22:56:32 2024

@author: jonathan
"""

from typing import Dict
import sys

from qtpy import QtCore, QtWidgets

from .icons import icons
from .GUI_instances import clearPreferences
from ..utilities import boolean
from ..config import (check_autolab_config, change_autolab_config,
                      check_plotter_config, change_plotter_config,
                      load_config, modify_config, set_temp_folder,
                      add_extra_driver_path)
from ..drivers import update_drivers_paths


class PreferencesWindow(QtWidgets.QMainWindow):

    def __init__(self, parent: QtWidgets.QMainWindow = None):

        super().__init__()
        self.mainGui = parent
        self.setWindowTitle('AUTOLAB - Preferences')
        self.setWindowIcon(icons['preference'])

        self.setFocus()
        self.activateWindow()
        self.statusBar = self.statusBar()

        self.init_ui()

        self.adjustSize()

        self.resize(500, 670)

    def init_ui(self):
        # Update config if needed
        check_autolab_config()
        check_plotter_config()

        autolab_config = load_config('autolab_config')
        plotter_config = load_config('plotter_config')

        centralWidget = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout(centralWidget)
        layout.setContentsMargins(0,0,0,0)
        layout.setSpacing(0)
        self.setCentralWidget(centralWidget)

        scrollAutolab = QtWidgets.QScrollArea()
        scrollAutolab.setWidgetResizable(True)
        frameAutolab = QtWidgets.QFrame()
        scrollAutolab.setWidget(frameAutolab)
        layoutAutolab = QtWidgets.QVBoxLayout(frameAutolab)
        layoutAutolab.setAlignment(QtCore.Qt.AlignTop)

        scrollPlotter = QtWidgets.QScrollArea()
        scrollPlotter.setWidgetResizable(True)
        framePlotter = QtWidgets.QFrame()
        scrollPlotter.setWidget(framePlotter)
        layoutPlotter = QtWidgets.QVBoxLayout(framePlotter)
        layoutPlotter.setAlignment(QtCore.Qt.AlignTop)

        tab = QtWidgets.QTabWidget(self)
        tab.addTab(scrollAutolab, 'Autolab')
        tab.addTab(scrollPlotter, 'Plotter')

        layout.addWidget(tab)

        # Create a frame for each main key in the dictionary
        self.inputs_autolab = {}
        self.inputs_plotter = {}

        ### Autolab
        ## GUI
        main_key = 'GUI'
        group_box = QtWidgets.QGroupBox(main_key)
        layoutAutolab.addWidget(group_box)
        group_layout = QtWidgets.QFormLayout(group_box)
        self.inputs_autolab[main_key] = {}

        sub_key = 'qt_api'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QComboBox()
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select the GUI Qt binding')
        input_widget.addItems(['default', 'pyqt5', 'pyqt6', 'pyside2', 'pyside6'])
        index = input_widget.findText(saved_value)
        input_widget.setCurrentIndex(index)
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(QtWidgets.QLabel(sub_key), input_widget)

        sub_key = 'theme'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QComboBox()
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select the GUI theme')
        input_widget.addItems(['default', 'dark'])
        index = input_widget.findText(saved_value)
        input_widget.setCurrentIndex(index)
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(QtWidgets.QLabel(sub_key), input_widget)

        sub_key = 'font_size'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QSpinBox()
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select the GUI text font size')
        input_widget.setValue(int(float(saved_value)))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(QtWidgets.QLabel(sub_key), input_widget)

        sub_key = 'image_background'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QLineEdit()
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select the plots background color (disabled if use dark theme)')
        input_widget.setText(str(saved_value))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(QtWidgets.QLabel(sub_key), input_widget)

        sub_key = 'image_foreground'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QLineEdit()
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select the plots foreground color (disabled if use dark theme)')
        input_widget.setText(str(saved_value))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(QtWidgets.QLabel(sub_key), input_widget)

        ## control_center
        main_key = 'control_center'
        group_box = QtWidgets.QGroupBox(main_key)
        layoutAutolab.addWidget(group_box)
        group_layout = QtWidgets.QFormLayout(group_box)
        self.inputs_autolab[main_key] = {}

        sub_key = 'precision'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QSpinBox()
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select the displayed precision for variables in the control panel')
        input_widget.setValue(int(float(saved_value)))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(QtWidgets.QLabel(sub_key), input_widget)

        sub_key = 'print'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QCheckBox(sub_key)
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select if print GUI information to console')
        input_widget.setChecked(boolean(saved_value))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(input_widget)

        sub_key = 'logger'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QCheckBox(sub_key)
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Activate a logger to display GUI information')
        input_widget.setChecked(boolean(saved_value))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(input_widget)

        sub_key = 'console'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QCheckBox(sub_key)
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Activate a console for debugging')
        input_widget.setChecked(boolean(saved_value))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(input_widget)

        ## monitor
        main_key = 'monitor'
        group_box = QtWidgets.QGroupBox(main_key)
        layoutAutolab.addWidget(group_box)
        group_layout = QtWidgets.QFormLayout(group_box)
        self.inputs_autolab[main_key] = {}

        sub_key = 'precision'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QSpinBox()
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select the displayed precision for variables in monitors')
        input_widget.setValue(int(float(saved_value)))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(QtWidgets.QLabel(sub_key), input_widget)

        sub_key = 'save_figure'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QCheckBox(sub_key)
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select if save figure image when saving monitor data')
        input_widget.setChecked(boolean(saved_value))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(input_widget)

        ## scanner
        main_key = 'scanner'
        group_box = QtWidgets.QGroupBox(main_key)
        layoutAutolab.addWidget(group_box)
        group_layout = QtWidgets.QFormLayout(group_box)
        self.inputs_autolab[main_key] = {}

        sub_key = 'precision'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QSpinBox()
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select the displayed precision for variables in the scanner')
        input_widget.setValue(int(float(saved_value)))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(QtWidgets.QLabel(sub_key), input_widget)

        sub_key = 'save_config'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QCheckBox(sub_key)
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select if save config file when saving scanner data')
        input_widget.setChecked(boolean(saved_value))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(input_widget)

        sub_key = 'save_figure'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QCheckBox(sub_key)
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select if save figure image when saving scanner data')
        input_widget.setChecked(boolean(saved_value))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(input_widget)

        sub_key = 'save_temp'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QCheckBox(sub_key)
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select if save temporary data. Should not be disable!')
        input_widget.setChecked(boolean(saved_value))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(input_widget)

        sub_key = 'ask_close'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QCheckBox(sub_key)
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Set whether a warning message about unsaved data should be displayed when the scanner is closed.')
        input_widget.setChecked(boolean(saved_value))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(input_widget)

        ## directories
        main_key = 'directories'
        group_box = QtWidgets.QGroupBox(main_key)
        layoutAutolab.addWidget(group_box)
        group_layout = QtWidgets.QFormLayout(group_box)
        self.inputs_autolab[main_key] = {}

        sub_key = 'temp_folder'
        saved_value = autolab_config[main_key][sub_key]
        input_widget = QtWidgets.QLineEdit()
        input_widget.setToolTip("Select the temporary folder. default correspond to os.environ['TEMP']")
        input_widget.setText(str(saved_value))
        self.inputs_autolab[main_key][sub_key] = input_widget
        group_layout.addRow(QtWidgets.QLabel(sub_key), input_widget)

        ## extra_driver_path
        main_key = 'extra_driver_path'
        group_box = QtWidgets.QFrame()
        group_layout = QtWidgets.QVBoxLayout(group_box)
        group_layout.setContentsMargins(0,0,0,0)
        layoutAutolab.addWidget(group_box)
        group_box_plugin = QtWidgets.QGroupBox(main_key)
        group_layout.addWidget(group_box_plugin)
        group_box.setToolTip('Add extra driver location')
        folder_layout = QtWidgets.QFormLayout(group_box_plugin)
        self.inputs_autolab[main_key] = {}

        addPluginButton = QtWidgets.QPushButton('Add driver folder')
        addPluginButton.setIcon(icons['add'])
        addPluginButton.clicked.connect(lambda: self.addOptionClicked(
            folder_layout, self.inputs_autolab, 'extra_driver_path',
            'onedrive',
            r'C:\Users\username\OneDrive\my_drivers'))
        group_layout.addWidget(addPluginButton)

        for sub_key, saved_value in autolab_config[main_key].items():
            self.addOptionClicked(
                folder_layout, self.inputs_autolab, main_key, sub_key, saved_value)

        ## extra_driver_url_repo
        main_key = 'extra_driver_url_repo'
        group_box = QtWidgets.QFrame()
        group_layout = QtWidgets.QVBoxLayout(group_box)
        group_layout.setContentsMargins(0,0,0,0)
        layoutAutolab.addWidget(group_box)
        group_box_plugin = QtWidgets.QGroupBox(main_key)
        group_layout.addWidget(group_box_plugin)
        group_box.setToolTip('Add extra url to download drivers from')
        url_layout = QtWidgets.QFormLayout(group_box_plugin)
        self.inputs_autolab[main_key] = {}

        addPluginButton = QtWidgets.QPushButton('Add driver url')
        addPluginButton.setIcon(icons['add'])
        addPluginButton.clicked.connect(lambda: self.addOptionClicked(
            url_layout, self.inputs_autolab, 'extra_driver_url_repo',
            r'C:\Users\username\OneDrive\my_drivers',
            r'https://github.com/my_repo/my_drivers'))
        group_layout.addWidget(addPluginButton)

        for sub_key, saved_value in autolab_config[main_key].items():
            self.addOptionClicked(
                url_layout, self.inputs_autolab, main_key, sub_key, saved_value)

        ### Plotter
        ## plugin
        main_key = 'plugin'
        group_box = QtWidgets.QFrame()
        group_layout = QtWidgets.QVBoxLayout(group_box)
        group_layout.setContentsMargins(0,0,0,0)
        group_layout.setSpacing(0)
        layoutPlotter.addWidget(group_box)
        group_box_plugin = QtWidgets.QGroupBox(main_key)
        group_layout.addWidget(group_box_plugin)
        group_box_plugin.setToolTip('Add plugins to plotter')
        plugin_layout = QtWidgets.QFormLayout(group_box_plugin)
        self.inputs_plotter[main_key] = {}

        addPluginButton = QtWidgets.QPushButton('Add plugin')
        addPluginButton.setIcon(icons['add'])
        addPluginButton.clicked.connect(lambda: self.addOptionClicked(
            plugin_layout, self.inputs_plotter, 'plugin', 'plugin', 'plotter'))
        group_layout.addWidget(addPluginButton)

        for sub_key, saved_value in plotter_config[main_key].items():
            self.addOptionClicked(
                plugin_layout, self.inputs_plotter, main_key, sub_key, saved_value)
            # To disable plotter modification
            input_key_widget, input_widget = self.inputs_plotter[main_key][sub_key]
            if sub_key == 'plotter':
                input_key_widget.setEnabled(False)
                input_widget.setEnabled(False)

        ## device
        main_key = 'device'
        group_box = QtWidgets.QGroupBox(main_key)
        layoutPlotter.addWidget(group_box)
        group_layout = QtWidgets.QFormLayout(group_box)
        self.inputs_plotter[main_key] = {}

        sub_key = 'address'
        saved_value = plotter_config[main_key][sub_key]
        input_widget = QtWidgets.QLineEdit()
        input_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_widget.setToolTip('Select the address of a device variable to be captured by the plotter.')
        input_widget.setText(str(saved_value))
        self.inputs_plotter[main_key][sub_key] = input_widget
        group_layout.addRow(QtWidgets.QLabel(sub_key), input_widget)

        ### Buttons
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok
                | QtWidgets.QDialogButtonBox.Cancel
                | QtWidgets.QDialogButtonBox.Apply,
            self)

        layout.addWidget(button_box)
        button_box.setContentsMargins(6,6,6,6)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        apply_button = button_box.button(QtWidgets.QDialogButtonBox.Apply)
        apply_button.clicked.connect(self.apply)

    def addOptionClicked(self, main_layout: QtWidgets.QLayout,
                         main_inputs: dict, main_key: str, sub_key: str, val: str):
        """ Add new option layout """
        basename = sub_key
        names = main_inputs[main_key].keys()
        compt = 0
        while True:
            if sub_key in names:
                compt += 1
                sub_key = basename + '_' + str(compt)
            else:
                break

        layout = QtWidgets.QHBoxLayout()
        input_key_widget = QtWidgets.QLineEdit()
        input_key_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        input_key_widget.setText(sub_key)
        layout.addWidget(input_key_widget)
        input_widget = QtWidgets.QLineEdit()
        input_widget.setText(val)
        layout.addWidget(input_widget)
        button_widget = QtWidgets.QPushButton()
        button_widget.setIcon(icons['remove'])
        button_widget.clicked.connect(lambda: self.removeOptionClicked(
            layout, main_inputs, main_key, sub_key))
        layout.addWidget(button_widget)

        main_inputs[main_key][sub_key] = (input_key_widget, input_widget)
        main_layout.addRow(layout)

    def removeOptionClicked(self, layout: QtWidgets.QLayout,
                            main_inputs: dict, main_key: str,
                            sub_key: str):
        """ Remove optional argument layout """
        for j in reversed(range(layout.count())):
            layout.itemAt(j).widget().setParent(None)
        layout.setParent(None)
        main_inputs[main_key].pop(sub_key)

    def accept(self):
        try:
            self.generate_new_configs()
        except Exception as e:
            self.setStatus(f'Config error: {e}', 10000, False)
        else:
            self.close()

    def apply(self):
        try:
            self.generate_new_configs()
        except Exception as e:
            self.setStatus(f'Config error: {e}', 10000, False)

    def reject(self):
        self.close()

    @staticmethod
    def generate_new_dict(input_widgets: Dict[str, QtWidgets.QWidget]) -> dict:

        new_dict = {}
        for main_key, sub_dict in input_widgets.items():
            new_dict[main_key] = {}
            for sub_key, widget in sub_dict.items():
                if isinstance(widget, QtWidgets.QSpinBox):
                    new_dict[main_key][sub_key] = widget.value()
                elif isinstance(widget, QtWidgets.QDoubleSpinBox):
                    new_dict[main_key][sub_key] = widget.value()
                elif isinstance(widget, QtWidgets.QCheckBox):
                    new_dict[main_key][sub_key] = widget.isChecked()
                elif isinstance(widget, QtWidgets.QComboBox):
                    new_dict[main_key][sub_key] = widget.currentText()
                elif isinstance(widget, tuple):
                    key_widget, value_widget = widget
                    new_dict[main_key][key_widget.text()] = value_widget.text()
                else:
                    new_dict[main_key][sub_key] = widget.text()

        return new_dict

    def generate_new_configs(self):
        new_autolab_dict = self.generate_new_dict(self.inputs_autolab)
        new_autolab_config = modify_config('autolab_config', new_autolab_dict)
        autolab_config = load_config('autolab_config')

        if new_autolab_config != autolab_config:
            change_autolab_config(new_autolab_config)

        new_plotter_dict = self.generate_new_dict(self.inputs_plotter)
        new_plotter_config = modify_config('plotter_config', new_plotter_dict)
        plotter_config = load_config('plotter_config')

        if new_plotter_config != plotter_config:
            change_plotter_config(new_plotter_config)

        if (new_autolab_config['directories']['temp_folder']
                != autolab_config['directories']['temp_folder']):
            set_temp_folder()

        if (new_autolab_config['extra_driver_path']
                != autolab_config['extra_driver_path']):
            add_extra_driver_path()
            update_drivers_paths()

        if (new_autolab_config['control_center']['logger'] != autolab_config['control_center']['logger']
                and hasattr(self.mainGui, 'activate_logger')):
            self.mainGui.activate_logger(new_autolab_config['control_center']['logger'])

        if (new_autolab_config['control_center']['console'] != autolab_config['control_center']['console']
                and hasattr(self.mainGui, 'activate_console')):
            self.mainGui.activate_console(new_autolab_config['control_center']['console'])

        if (new_autolab_config['GUI']['qt_api'] != autolab_config['GUI']['qt_api']
                or new_autolab_config['GUI']['theme'] != autolab_config['GUI']['theme']
                or new_autolab_config['GUI']['font_size'] != autolab_config['GUI']['font_size']
                or new_autolab_config['GUI']['image_background'] != autolab_config['GUI']['image_background']
                or new_autolab_config['GUI']['image_background'] != autolab_config['GUI']['image_background']
                ):
            QtWidgets.QMessageBox.information(
                self,
                'Information',
                'One or more of the settings you changed requires a restart to be applied',
                QtWidgets.QMessageBox.Ok
                )

    def setStatus(self, message: str, timeout: int = 0, stdout: bool = True):
        """ Modify the message displayed in the status bar and add error message to logger """
        self.statusBar.showMessage(message, timeout)
        if not stdout: print(message, file=sys.stderr)

    def closeEvent(self, event):
        """ Does some steps before the window is really killed """
        clearPreferences()

        if not self.mainGui:
            QtWidgets.QApplication.quit()  # close the app
